@@ -34,6 +34,7 @@
            "FUNCTION-ARGLIST"
            "FUNCTION-LAMBDA-LIST"
            "FUNCTION-TYPE"
+           "METHOD-COMBINATION-LAMBDA-LIST"
            "DEFTYPE-LAMBDA-LIST"
            "VALID-FUNCTION-NAME-P"
            "FIND-DEFINITION-SOURCE"
@@ -99,7 +100,8 @@ include the pathname of the file and the position of the definition."
 (declaim (ftype (sfunction (t debug-info) debug-function) debug-info-debug-function))
 (defun debug-info-debug-function (function debug-info)
   (sb-di::compiled-debug-fun-from-pc debug-info
-                                     (sb-di::function-start-pc-offset function)))
+                                     (sb-di::function-start-pc-offset function)
+                                     t))
 
 (defun valid-function-name-p (name)
   "True if NAME denotes a valid function name, ie. one that can be passed to
@@ -300,11 +302,11 @@ If an unsupported TYPE is requested, the function will return NIL.
                    (eq (car name) 'setf))
           (setf name (cadr name)))
         (let ((expander (info :setf :expander name)))
-          (when expander
-            (find-definition-source
-             (cond ((symbolp expander) (symbol-function expander))
-                   ((listp expander) (cdr expander))
-                   (t expander))))))
+          (cond ((typep expander '(cons symbol))
+                 (translate-source-location (cddr expander)))
+                (expander
+                 (find-definition-source
+                  (if (listp expander) (cdr expander) expander))))))
        ((:structure)
         (let ((class (get-class name)))
           (if class
@@ -353,6 +355,7 @@ If an unsupported TYPE is requested, the function will return NIL.
                             (sb-c:fun-info-ltn-annotate . sb-c:ltn-annotate)
                             (sb-c:fun-info-optimizer . sb-c:optimizer)
                             (sb-c:fun-info-ir2-convert . sb-c:ir2-convert)
+                            (sb-c::fun-info-ir2-hook . sb-c::ir2-hook)
                             (sb-c::fun-info-stack-allocate-result
                              . sb-c::stack-allocate-result)
                             (sb-c::fun-info-constraint-propagate
@@ -459,13 +462,14 @@ If an unsupported TYPE is requested, the function will return NIL.
   (let* ((debug-info (function-debug-info function))
          (debug-source (debug-info-source debug-info))
          (debug-fun (debug-info-debug-function function debug-info))
-         (tlf (sb-c::compiled-debug-info-tlf-number debug-info)))
+         (tlf (if debug-fun (sb-c::compiled-debug-fun-tlf-number debug-fun))))
     (make-definition-source
      :pathname
      (when (stringp (debug-source-namestring debug-source))
        (parse-namestring (debug-source-namestring debug-source)))
      :character-offset
-     (sb-c::compiled-debug-info-char-offset debug-info)
+     (if tlf
+         (elt (sb-c::debug-source-start-positions debug-source) tlf))
      :form-path (if tlf (list tlf))
      :form-number (sb-c::compiled-debug-fun-form-number debug-fun)
      :file-write-date (debug-source-created debug-source)
@@ -492,11 +496,13 @@ If an unsupported TYPE is requested, the function will return NIL.
   (function-lambda-list function))
 
 (defun function-lambda-list (function)
-  "Describe the lambda list for the extended function designator FUNCTION.
+  "Return the lambda list for the extended function designator FUNCTION.
 Works for special-operators, macros, simple functions, interpreted functions,
 and generic functions. Signals an error if FUNCTION is not a valid extended
-function designator."
-  ;; FIXME: sink this logic into SB-KERNEL:%FUN-LAMBDA-LIST and just call that?
+function designator.
+
+If the function does not have a lambda list (compiled with debug 0),
+then two values are returned: (values nil t)"
   (cond ((and (symbolp function) (special-operator-p function))
          (function-lambda-list (info :function :ir1-convert function)))
         ((valid-function-name-p function)
@@ -506,7 +512,10 @@ function designator."
         ((typep function 'generic-function)
          (sb-pcl::generic-function-pretty-arglist function))
         (t
-         (%fun-lambda-list function))))
+         (let ((raw-result (%fun-lambda-list function)))
+           (if (eq raw-result :unknown)
+               (values nil t)
+               (values raw-result nil))))))
 
 (defun deftype-lambda-list (typespec-operator)
   "Returns the lambda list of TYPESPEC-OPERATOR as first return
@@ -521,6 +530,18 @@ value."
         (values (%fun-lambda-list f) t)
         (values nil nil))))
 
+(defun method-combination-lambda-list (method-combination)
+  "Return the lambda-list of METHOD-COMBINATION designator.
+METHOD-COMBINATION can be a method combination object,
+or a method combination name."
+  (let* ((name (etypecase method-combination
+                 (symbol method-combination)
+                 (method-combination
+                  (sb-pcl::method-combination-type-name method-combination))))
+         (info (or (gethash name sb-pcl::**method-combinations**)
+                   (error "~S: no such method combination." name))))
+    (sb-pcl::method-combination-info-lambda-list info)))
+
 (defun function-type (function-designator)
   "Returns the ftype of FUNCTION-DESIGNATOR, or NIL."
   (etypecase function-designator
@@ -541,7 +562,7 @@ value."
            ;; because it contains more accurate information e.g. for
            ;; struct-accessors.
            (function-type name)
-           (sb-impl::%fun-type function-designator))))))
+           (sb-impl::%fun-ftype function-designator))))))
 
 ;;;; find callers/callees, liberated from Helmut Eller's code in SLIME
 
@@ -597,7 +618,21 @@ value."
                 ;; entry.
                 (setf (definition-source-form-number source-location)
                       xref-form-number)
-                (push (cons name source-location) result))))
+                (let ((name (cond ((sb-c::transform-p name)
+                                   (let ((fun-name (%fun-name fun)))
+                                     (append (if (consp fun-name)
+                                                 fun-name
+                                                 (list fun-name))
+                                             (let* ((type (sb-c::transform-type name))
+                                                    (type-spec (type-specifier type)))
+                                               (and (sb-kernel:fun-type-p type)
+                                                    (list (second type-spec)))))))
+                                  ((sb-c::vop-info-p name)
+                                   (list 'sb-c:define-vop
+                                         (sb-c::vop-info-name name)))
+                                  (t
+                                   name))))
+                  (push (cons name source-location) result)))))
           xrefs))))
     result))
 
@@ -718,6 +753,14 @@ Experimental.
 
 ;;;; ALLOCATION INTROSPECTION
 
+(eval-when (:compile-toplevel :execute)
+  (defmacro pinnedp (addr)
+    `(eql (sb-alien:alien-funcall
+           (sb-alien:extern-alien "sb_introspect_pinnedp"
+                                  (function sb-alien:int sb-alien:unsigned))
+           ,addr)
+          1)))
+
 (defun allocation-information (object)
   "Returns information about the allocation of OBJECT. Primary return value
 indicates the general type of allocation: :IMMEDIATE, :HEAP, :STACK,
@@ -784,27 +827,15 @@ Experimental: interface subject to change."
                          ;; bits are packed in the opposite order. And thankfully,
                          ;; this fix seems not to depend on whether the numbering
                          ;; scheme is MSB 0 or LSB 0, afaict.
-                         (let* ((index (sb-vm:find-page-index
+                         (let* ((wp (page-protected-p object))
+                                (index (sb-vm:find-page-index
                                         (get-lisp-obj-address object)))
-                                (flags (sb-alien:slot page 'sb-vm::flags))
-                                .
-                                ;; The unused WP-CLR is for ease of counting
-                                #+big-endian
-                                ((type      (ldb (byte 5 3) flags))
-                                 (wp        (logbitp 2 flags))
-                                 (wp-clr    (logbitp 1 flags))
-                                 (dontmove  (logbitp 0 flags)))
-                                #+little-endian
-                                ((type      (ldb (byte 5 0) flags))
-                                 (wp        (logbitp 5 flags))
-                                 (wp-clr    (logbitp 6 flags))
-                                 (dontmove  (logbitp 7 flags))))
-                           (declare (ignore wp-clr))
+                                (type (sb-alien:slot page 'sb-vm::flags)))
                            (list :space space
                                  :generation (sb-alien:slot page 'sb-vm::gen)
                                  :write-protected wp
-                                 :boxed (logbitp 0 type)
-                                 :pinned dontmove
+                                 :boxed (> (logand type #xf) 1)
+                                 :pinned (pinnedp (get-lisp-obj-address object))
                                  :large (logbitp 4 type)
                                  :page index)))
                        (list :space space))
@@ -853,6 +884,7 @@ Experimental: interface subject to change."
                         (not (xset-member-p part seen)))
                  (add-to-xset part seen)
                  (funcall fun part))))
+      (declare (truly-dynamic-extent #'call))
       (when ext
         (multiple-value-bind (value foundp)
             (let ((table sb-pcl::*eql-specializer-table*))
@@ -1005,7 +1037,7 @@ Experimental: interface subject to change."
                                    (= this-bin-size (+ prev-bin-size 2)))
                                this-bin-size))))))))
 
-(defun largest-objects (&key (threshold #+gencgc sb-vm:gencgc-card-bytes
+(defun largest-objects (&key (threshold #+gencgc sb-vm:gencgc-page-bytes
                                         #-gencgc sb-c:+backend-page-bytes+)
                              (sort :size))
   (declare (type (member :address :size) sort))
