@@ -232,9 +232,10 @@
   (define-call-internally ioctl-without-arg "ioctl" int minusp
                           (fd file-descriptor) (cmd unsigned-long))
   (define-call-internally ioctl-with-int-arg "ioctl" int minusp
-                          (fd file-descriptor) (cmd unsigned-long) (arg int))
+                          (fd file-descriptor) (cmd unsigned-long) &optional (arg int))
   (define-call-internally ioctl-with-pointer-arg "ioctl" int minusp
                           (fd file-descriptor) (cmd unsigned-long)
+                          &optional
                           (arg alien-pointer-to-anything-or-nil))
   (define-entry-point "ioctl" (fd cmd &optional (arg nil argp))
     (if argp
@@ -547,8 +548,8 @@ not supported."
    (passwd :initarg :passwd :accessor group-passwd)
    (gid :initarg :gid :accessor group-gid)))
 
-(defmacro define-obj-call (name arg type conv)
-  #-(or win32 android)
+#-(or win32 android)
+(macrolet ((define-obj-call (name arg type conv)
   ;; FIXME: this isn't the documented way of doing this, surely?
   (let ((lisp-name (intern (string-upcase name) :sb-posix)))
     `(progn
@@ -558,7 +559,7 @@ not supported."
         (let ((r (alien-funcall (extern-alien ,name ,type) ,arg)))
           (if (null-alien r)
               nil
-              (,conv r)))))))
+              (,conv r))))))))
 
 (define-obj-call "getpwnam" login-name (function (* alien-passwd) (c-string :not-null t))
                  alien-to-passwd)
@@ -568,7 +569,7 @@ not supported."
                  alien-to-group)
 (define-obj-call "getgrgid" gid (function (* alien-group) gid-t)
                  alien-to-group)
-
+) ; end MACROLET
 
 #-win32
 (define-protocol-class timeval alien-timeval ()
