@@ -9,14 +9,14 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!IMPL")
+(in-package "SB-IMPL")
 
 
 ;;; ASCII
 
 (declaim (inline code->ascii-mapper))
 (defun code->ascii-mapper (code)
-  (declare (optimize speed (safety 0))
+  (declare (optimize speed #.*safety-0*)
            (type char-code code))
   (if (> code 127)
       nil
@@ -24,13 +24,13 @@
 
 (declaim (inline get-ascii-bytes))
 (defun get-ascii-bytes (string pos)
-  (declare (optimize speed (safety 0))
+  (declare (optimize speed #.*safety-0*)
            (type simple-string string)
            (type array-range pos))
   (get-latin-bytes #'code->ascii-mapper :ascii string pos))
 
 (defun string->ascii (string sstart send null-padding)
-  (declare (optimize speed (safety 0))
+  (declare (optimize speed #.*safety-0*)
            (type simple-string string)
            (type array-range sstart send))
   (values (string->latin% string sstart send #'get-ascii-bytes null-padding)))
@@ -74,13 +74,13 @@
 
 (declaim (inline get-latin1-bytes))
 (defun get-latin1-bytes (string pos)
-  (declare (optimize speed (safety 0))
+  (declare (optimize speed #.*safety-0*)
            (type simple-string string)
            (type array-range pos))
   (get-latin-bytes #'identity :latin-1 string pos))
 
 (defun string->latin1 (string sstart send null-padding)
-  (declare (optimize speed (safety 0))
+  (declare (optimize speed #.*safety-0*)
            (type simple-string string)
            (type array-range sstart send))
   (values (string->latin% string sstart send #'get-latin1-bytes null-padding)))
@@ -117,8 +117,8 @@
 
 (declaim (inline char-len-as-utf8))
 (defun char-len-as-utf8 (code)
-  (declare (optimize speed (safety 0))
-           (type (integer 0 (#.sb!xc:char-code-limit)) code))
+  (declare (optimize speed #.*safety-0*)
+           (type (integer 0 (#.char-code-limit)) code))
   (cond ((< code 0) (bug "can't happen"))
         ((< code #x80) 1)
         ((< code #x800) 2)
@@ -127,7 +127,7 @@
         (t (bug "can't happen"))))
 
 (defun string->utf8 (string sstart send null-padding)
-  (declare (optimize (speed 3) (safety 0))
+  (declare (optimize (speed 3) #.*safety-0*)
            (type simple-string string)
            (type (integer 0 1) null-padding)
            (type array-range sstart send))
@@ -198,18 +198,13 @@
                                         (unless (zerop null-padding)
                                           (vector-push-extend 0 new-array))
                                         (copy-seq new-array)))))))))))
-      #!+sb-unicode
+      #+sb-unicode
       ((simple-array base-char (*))
        ;; On unicode builds BASE-STRINGs are limited to ASCII range,
        ;; so we can take a fast path -- and get benefit of the element
        ;; type information. On non-unicode build BASE-CHAR ==
        ;; CHARACTER, handled above.
-       (ascii-bash))
-      ((simple-array nil (*))
-       (if (= send sstart)
-           (make-array null-padding :element-type '(unsigned-byte 8))
-           ;; Just get the error...
-           (aref string sstart))))))
+       (ascii-bash)))))
 
 ;;; from UTF-8
 
@@ -218,11 +213,11 @@
     `(progn
       ;;(declaim (inline ,name))
       (let ((lexically-max
-             (string->utf8 (string (code-char ,(1- sb!xc:char-code-limit)))
+             (string->utf8 (string (code-char ,(1- char-code-limit)))
                            0 1 0)))
-        (declare (type (simple-array (unsigned-byte 8) (#!+sb-unicode 4 #!-sb-unicode 2)) lexically-max))
+        (declare (type (simple-array (unsigned-byte 8) (#+sb-unicode 4 #-sb-unicode 2)) lexically-max))
         (defun ,name (array pos end)
-          (declare (optimize speed (safety 0))
+          (declare (optimize speed #.*safety-0*)
                    (type ,type array)
                    (type array-range pos end))
           ;; returns the number of bytes consumed and nil if it's a
@@ -288,16 +283,16 @@
                                         t)
                                        ((< (aref lexically-max 1) (,accessor array (+ pos 1)))
                                         nil)
-                                       #!+sb-unicode
+                                       #+sb-unicode
                                        ((> (aref lexically-max 1) (,accessor array (+ pos 1)))
                                         t)
-                                       #!+sb-unicode
+                                       #+sb-unicode
                                        ((< (aref lexically-max 2) (,accessor array (+ pos 2)))
                                         nil)
-                                       #!+sb-unicode
+                                       #+sb-unicode
                                        ((> (aref lexically-max 2) (,accessor array (+ pos 2)))
                                         t)
-                                       #!+sb-unicode
+                                       #+sb-unicode
                                        ((< (aref lexically-max 3) (,accessor array (+ pos 3)))
                                         nil)
                                        (t t))))
@@ -312,13 +307,13 @@
                 (cond ((eql maybe-len 1)
                        (values 1 nil))
                       ((and (preliminary-ok-for-length maybe-len 2)
-                            #!-sb-unicode (character-below-char-code-limit-p))
+                            #-sb-unicode (character-below-char-code-limit-p))
                        (values 2 nil))
                       ((and (preliminary-ok-for-length maybe-len 3)
-                            #!-sb-unicode (not (setf reject-reason 'character-out-of-range)))
+                            #-sb-unicode (not (setf reject-reason 'character-out-of-range)))
                        (values 3 nil))
                       ((and (preliminary-ok-for-length maybe-len 4)
-                            #!-sb-unicode (not (setf reject-reason 'character-out-of-range))
+                            #-sb-unicode (not (setf reject-reason 'character-out-of-range))
                             (character-below-char-code-limit-p))
                        (values 4 nil))
                       (t
@@ -339,7 +334,7 @@
     `(progn
       (declaim (inline ,name))
       (defun ,name (array pos bytes)
-        (declare (optimize speed (safety 0))
+        (declare (optimize speed #.*safety-0*)
                  (type ,type array)
                  (type array-range pos)
                  (type (integer 1 4) bytes))
@@ -363,7 +358,7 @@
   (let ((name (make-od-name 'utf8->string accessor)))
     `(progn
       (defun ,name (array astart aend)
-        (declare (optimize speed (safety 0))
+        (declare (optimize speed #.*safety-0*)
                  (type ,type array)
                  (type array-range astart aend))
         (let ((string (make-array 0 :adjustable t :fill-pointer 0 :element-type 'character)))
@@ -382,9 +377,134 @@
           (coerce string 'simple-string))))))
 (instantiate-octets-definition define-utf8->string)
 
+#+(and sb-unicode 64-bit little-endian)
+(defun copy-utf8-bytes-to-character-string (requested total-copied start buffer ibuf)
+  (declare (type index start requested total-copied)
+           (optimize speed (safety 0)))
+  (with-pinned-objects (buffer)
+    (let* ((head (buffer-head ibuf))
+           (tail (buffer-tail ibuf))
+           (sap (buffer-sap ibuf))
+           (left (- requested total-copied))
+           (result-characters (logand left (- sb-vm:n-word-bytes)))
+           (buffer-bytes (logand (- tail head) (- sb-vm:n-word-bytes)))
+           (n (min result-characters buffer-bytes))
+           (repeat (ldb (byte sb-vm:n-word-bits 0) #x0101010101010101))
+           (ascii-mask (* 128 repeat))
+           (buffer-sap (vector-sap buffer))
+           (buffer-start (truly-the fixnum (* (+ start total-copied) 4))))
+      (declare (optimize sb-c::preserve-single-use-debug-variables
+                         sb-c::preserve-constants))
+      (loop for ibuf-offset from head below (+ head n) by sb-vm:n-word-bytes
+            for buffer-offset from buffer-start by (* 4 sb-vm:n-word-bytes)
+            do
+            (let ((word (sb-sys:sap-ref-word sap ibuf-offset)))
+              (when (logtest word ascii-mask)
+                (setf head ibuf-offset)
+                (incf total-copied
+                      (truncate (- buffer-offset buffer-start) 4))
+                (return))
+              (setf (sap-ref-word buffer-sap buffer-offset)
+                    (dpb (ldb (byte 8 8) word)
+                         (byte 8 32)
+                         (ldb (byte 8 0) word))
+                    (sap-ref-word buffer-sap (+ buffer-offset 8))
+                    (dpb (ldb (byte 8 24) word)
+                         (byte 8 32)
+                         (ldb (byte 8 16) word))
+                    (sap-ref-word buffer-sap (+ buffer-offset 16))
+                    (dpb (ldb (byte 8 40) word)
+                         (byte 8 32)
+                         (ldb (byte 8 32) word))
+                    (sap-ref-word buffer-sap (+ buffer-offset 24))
+                    (dpb (ldb (byte 8 56) word)
+                         (byte 8 32)
+                         (ldb (byte 8 48) word))))
+            finally (incf head n)
+                    (incf total-copied n))
+      (setf (buffer-head ibuf) head)
+      total-copied)))
+
+(defun fd-stream-read-n-characters/utf-8 (stream buffer start requested eof-error-p &aux (total-copied 0))
+  (declare (type fd-stream stream)
+           (type index start requested total-copied)
+           (type (simple-array character (#.+ansi-stream-in-buffer-length+)) buffer))
+  (when (fd-stream-eof-forced-p stream)
+    (setf (fd-stream-eof-forced-p stream) nil)
+    (return-from fd-stream-read-n-characters/utf-8 0))
+  (do ((instead (fd-stream-instead stream)))
+      ((= (fill-pointer instead) 0)
+       (setf (fd-stream-listen stream) nil))
+    (setf (aref buffer (+ start total-copied)) (vector-pop instead))
+    (incf total-copied)
+    (when (= requested total-copied)
+      (when (= (fill-pointer instead) 0)
+        (setf (fd-stream-listen stream) nil))
+      (return-from fd-stream-read-n-characters/utf-8 total-copied)))
+  #+(and sb-unicode 64-bit little-endian)
+  (setf total-copied
+        (copy-utf8-bytes-to-character-string requested total-copied
+                                             start buffer (fd-stream-ibuf stream)))
+  (do ()
+      (())
+    (let* ((ibuf (fd-stream-ibuf stream))
+           (head (buffer-head ibuf))
+           (tail (buffer-tail ibuf))
+           (sap (buffer-sap ibuf))
+           (decode-break-reason nil))
+      (declare (type index head tail))
+      (do ((size nil nil))
+          ((or (= tail head) (= requested total-copied)))
+        (setf decode-break-reason
+              (block decode-break-reason
+                (when (< (- tail head) 1) (return))
+                (let ((byte (sap-ref-8 sap head)))
+                  (declare (ignorable byte))
+                  (setq size (cond ((< byte 128) 1) ((< byte 194) (return-from decode-break-reason 1)) ((< byte 224) 2) ((< byte 240) 3) (t 4)))
+                  (when (> size (- tail head)) (return))
+                  (setf (aref buffer (+ start total-copied))
+                        (code-char
+                         (ecase size
+                           (1 byte)
+                           (2
+                            (let ((byte2 (sap-ref-8 sap (1+ head))))
+                              (unless (<= 128 byte2 191) (return-from decode-break-reason 2))
+                              (dpb byte (byte 5 6) byte2)))
+                           (3
+                            (let ((byte2 (sap-ref-8 sap (1+ head))) (byte3 (sap-ref-8 sap (+ 2 head))))
+                              (unless
+                                  (and (<= 128 byte2 191) (<= 128 byte3 191) (or (/= byte 224) (<= 160 byte2 191))
+                                       (or (/= byte 237) (<= 128 byte2 159)))
+                                (return-from decode-break-reason 3))
+                              (dpb byte (byte 4 12) (dpb byte2 (byte 6 6) byte3))))
+                           (4
+                            (let ((byte2 (sap-ref-8 sap (1+ head))) (byte3 (sap-ref-8 sap (+ 2 head))) (byte4 (sap-ref-8 sap (+ 3 head))))
+                              (unless
+                                  (and (<= 128 byte2 191) (<= 128 byte3 191) (<= 128 byte4 191) (or (/= byte 240) (<= 144 byte2 191))
+                                       (or (/= byte 244) (<= 128 byte2 143)))
+                                (return-from decode-break-reason 4))
+                              (dpb byte (byte 3 18) (dpb byte2 (byte 6 12) (dpb byte3 (byte 6 6) byte4))))))))
+                  (incf total-copied)
+                  (incf head size))
+                nil))
+        (setf (buffer-head ibuf) head)
+        (when decode-break-reason
+          (when (plusp total-copied) (return-from fd-stream-read-n-characters/utf-8 total-copied))
+          (when (stream-decoding-error-and-handle stream decode-break-reason)
+            (if eof-error-p
+                (error 'end-of-file :stream stream)
+                (return-from fd-stream-read-n-characters/utf-8 total-copied)))
+          (return-from fd-stream-read-n-characters/utf-8 total-copied)))
+      (setf (buffer-head ibuf) head)
+      (cond ((plusp total-copied) (return total-copied))
+            ((or (eq decode-break-reason 'eof) (null (catch 'eof-input-catcher (refill-input-buffer stream))))
+             (if eof-error-p
+                 (error 'end-of-file :stream stream)
+                 (return total-copied)))))))
+
 (define-external-format/variable-width (:utf-8 :utf8) t
-  #!+sb-unicode (code-char #xfffd) #!-sb-unicode #\?
-  (let ((bits (char-code byte)))
+  #+sb-unicode (code-char #xfffd) #-sb-unicode #\?
+  (let ((bits (char-code |ch|)))
     (cond ((< bits #x80) 1)
           ((< bits #x800) 2)
           ((< bits #x10000) 3)
@@ -395,9 +515,9 @@
              (sap-ref-8 sap (+ 1 tail)) (logior #x80 (ldb (byte 6 0) bits))))
     (3 (when (<= #xd800 bits #xdfff)
          (external-format-encoding-error stream bits))
-       (setf (sap-ref-8 sap tail)       (logior #xe0 (ldb (byte 4 12) bits))
-             (sap-ref-8 sap (+ 1 tail)) (logior #x80 (ldb (byte 6 6) bits))
-             (sap-ref-8 sap (+ 2 tail)) (logior #x80 (ldb (byte 6 0) bits))))
+     (setf (sap-ref-8 sap tail)       (logior #xe0 (ldb (byte 4 12) bits))
+           (sap-ref-8 sap (+ 1 tail)) (logior #x80 (ldb (byte 6 6) bits))
+           (sap-ref-8 sap (+ 2 tail)) (logior #x80 (ldb (byte 6 0) bits))))
     (4 (setf (sap-ref-8 sap tail)       (logior #xf0 (ldb (byte 3 18) bits))
              (sap-ref-8 sap (+ 1 tail)) (logior #x80 (ldb (byte 6 12) bits))
              (sap-ref-8 sap (+ 2 tail)) (logior #x80 (ldb (byte 6 6) bits))
@@ -435,4 +555,5 @@
                               (dpb byte3 (byte 6 6) byte4)))))))
   utf8->string-aref
   string->utf8
-  #!+sb-unicode :base-string-direct-mapping #!+sb-unicode t)
+  #+sb-unicode :base-string-direct-mapping #+sb-unicode t
+  :fd-stream-read-n-characters fd-stream-read-n-characters/utf-8)
