@@ -9,10 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!IMPL")
-
-(defun time-reinit ()
-  (reinit-internal-real-time))
+(in-package "SB-IMPL")
 
 ;;; Implemented in unix.lisp and win32.lisp.
 (setf (documentation 'get-internal-real-time 'function)
@@ -145,7 +142,7 @@ format."
   (multiple-value-bind (seconds-west daylight)
       (if time-zone
           (values (* time-zone 60 60) nil)
-          (sb!unix::get-timezone (truncate-to-unix-range universal-time)))
+          (sb-unix::get-timezone (truncate-to-unix-range universal-time)))
     (declare (fixnum seconds-west))
     (multiple-value-bind (weeks secs)
         (truncate (+ (- universal-time seconds-west) seconds-offset)
@@ -193,14 +190,15 @@ format."
           (truncate years 100))
        (truncate (+ years 300) 400))))
 
-(defconstant +days-before-month+
+(defconstant-eqx +days-before-month+
   #.(let ((reversed-result nil)
           (sum 0))
       (push nil reversed-result)
       (dolist (days-in-month '(31 28 31 30 31 30 31 31 30 31 30 31))
         (push sum reversed-result)
         (incf sum days-in-month))
-      (coerce (nreverse reversed-result) 'simple-vector)))
+      (coerce (nreverse reversed-result) 'simple-vector))
+  #'equalp)
 
 (defun encode-universal-time (second minute hour date month year
                                      &optional time-zone)
@@ -232,27 +230,30 @@ format."
     (if time-zone
         (setf encoded-time (+ second (* (+ minute (* (+ hours time-zone) 60)) 60)))
         (let* ((secwest-guess
-                 (sb!unix::get-timezone
+                 (sb-unix::get-timezone
                   (truncate-to-unix-range (* hours 60 60))))
                (guess (+ second (* 60 (+ minute (* hours 60)))
                          secwest-guess))
                (secwest
-                 (sb!unix::get-timezone
+                 (sb-unix::get-timezone
                   (truncate-to-unix-range guess))))
           (setf encoded-time (+ guess (- secwest secwest-guess)))))
     encoded-time))
 
 ;;;; TIME
 
+(defvar *gc-real-time* 0
+  "Total real time spent doing garbage collection (as reported by
+GET-INTERNAL-REAL-TIME.) Initialized to zero on startup.")
 (defvar *gc-run-time* 0
   "Total CPU time spent doing garbage collection (as reported by
 GET-INTERNAL-RUN-TIME.) Initialized to zero on startup. It is safe to bind
 this to zero in order to measure GC time inside a certain section of code, but
 doing so may interfere with results reported by eg. TIME.")
-(declaim (type index *gc-run-time*))
+(declaim (type index *gc-run-time* *gc-real-time*))
 
 (defun print-time (&key real-time-ms user-run-time-us system-run-time-us
-                   gc-run-time-ms processor-cycles eval-calls
+                   gc-run-time-ms gc-real-time-ms processor-cycles eval-calls
                    lambdas-converted page-faults bytes-consed
                    aborted)
   (let ((total-run-time-us (+ user-run-time-us system-run-time-us))
@@ -265,6 +266,8 @@ doing so may interfere with results reported by eg. TIME.")
                          ~@<  ~@;~/sb-impl::format-milliseconds/ of real time~%~
                                  ~/sb-impl::format-microseconds/ of total run time ~
                                   (~@/sb-impl::format-microseconds/ user, ~@/sb-impl::format-microseconds/ system)~%~
+                                 ~[[ Real times consist of ~/sb-impl::format-milliseconds/ GC time, ~
+                                                       and ~/sb-impl::format-milliseconds/ non-GC time. ]~%~;~2*~]~
                                  ~[[ Run times consist of ~/sb-impl::format-milliseconds/ GC time, ~
                                                       and ~/sb-impl::format-milliseconds/ non-GC time. ]~%~;~2*~]~
                                  ~,2F% CPU~%~
@@ -278,12 +281,15 @@ doing so may interfere with results reported by eg. TIME.")
             total-run-time-us
             user-run-time-us
             system-run-time-us
+            (if (zerop gc-real-time-ms) 1 0)
+            gc-real-time-ms
+            (- real-time-ms gc-real-time-ms)
             (if (zerop gc-run-time-ms) 1 0)
             gc-run-time-ms
             ;; Round up so we don't mislead by saying 0.0 seconds of non-GC time...
             (- (ceiling total-run-time-us 1000) gc-run-time-ms)
             (if (zerop real-time-ms)
-                100.0
+                $100.0
                 (float (* 100 (/ (round total-run-time-us 1000) real-time-ms))))
             eval-calls
             lambdas-converted
@@ -317,16 +323,16 @@ normally during operations like SLEEP."
 
 (defun elapsed-cycles (h0 l0 h1 l1)
   (declare (ignorable h0 l0 h1 l1))
-  #!+cycle-counter
+  #+cycle-counter
   (+ (ash (- h1 h0) 32)
      (- l1 l0))
-  #!-cycle-counter
+  #-cycle-counter
   nil)
 (declaim (inline read-cycle-counter))
 (defun read-cycle-counter ()
-  #!+cycle-counter
-  (sb!vm::%read-cycle-counter)
-  #!-cycle-counter
+  #+cycle-counter
+  (sb-vm::%read-cycle-counter)
+  #-cycle-counter
   (values 0 0))
 
 ;;; This is so that we don't have to worry about the vagaries of
@@ -400,6 +406,9 @@ returns values returned by FUNCTION.
   :GC-RUN-TIME-MS
       GC run time in milliseconds (included in user and system run time.)
 
+  :GC-REAL-TIME-MS
+      GC real time in milliseconds.
+
   :PROCESSOR-CYCLES
       Approximate number of processor cycles used. (Omitted  if not supported on
       the platform -- currently available on x86 and x86-64 only.)
@@ -431,12 +440,8 @@ EXPERIMENTAL: Interface subject to change."
         old-page-faults
         new-page-faults
         real-time-overhead
-        run-utime-overhead
-        run-stime-overhead
-        page-faults-overhead
         old-bytes-consed
         new-bytes-consed
-        cons-overhead
         (fun (if (functionp function) function (fdefinition function))))
     (declare (function fun))
     ;; Calculate the overhead...
@@ -450,35 +455,33 @@ EXPERIMENTAL: Interface subject to change."
     (multiple-value-setq
         (new-run-utime new-run-stime new-page-faults new-bytes-consed)
       (time-get-sys-info))
-    (setq run-utime-overhead (- new-run-utime old-run-utime))
-    (setq run-stime-overhead (- new-run-stime old-run-stime))
-    (setq page-faults-overhead (- new-page-faults old-page-faults))
     (setq old-real-time (get-internal-real-time))
     (setq old-real-time (get-internal-real-time))
     (setq new-real-time (get-internal-real-time))
     (setq real-time-overhead (- new-real-time old-real-time))
-    (setq cons-overhead (- new-bytes-consed old-bytes-consed))
     ;; Now get the initial times.
     (multiple-value-setq
         (old-run-utime old-run-stime old-page-faults old-bytes-consed)
       (time-get-sys-info))
     (setq old-real-time (get-internal-real-time))
     (let ((start-gc-internal-run-time *gc-run-time*)
+          (start-gc-internal-real-time *gc-real-time*)
           (*eval-calls* 0)
-          (sb!c::*lambda-conversions* 0)
+          (sb-c::*lambda-conversions* 0)
           (aborted t))
-      (declare (special *eval-calls* sb!c::*lambda-conversions*))
       (multiple-value-bind (h0 l0) (read-cycle-counter)
         (unwind-protect
              (multiple-value-prog1 (apply fun arguments)
                (setf aborted nil))
           (multiple-value-bind (h1 l1) (read-cycle-counter)
-            (let ((stop-gc-internal-run-time *gc-run-time*))
+            (let ((stop-gc-internal-run-time *gc-run-time*)
+                  (stop-gc-internal-real-time *gc-real-time*))
               (multiple-value-setq
                   (new-run-utime new-run-stime new-page-faults new-bytes-consed)
                 (time-get-sys-info))
               (setq new-real-time (- (get-internal-real-time) real-time-overhead))
               (let* ((gc-internal-run-time (max (- stop-gc-internal-run-time start-gc-internal-run-time) 0))
+                     (gc-internal-real-time (max (- stop-gc-internal-real-time start-gc-internal-real-time) 0))
                      (real-time (max (- new-real-time old-real-time) 0))
                      (user-run-time (max (- new-run-utime old-run-utime) 0))
                      (system-run-time (max (- new-run-stime old-run-stime) 0))
@@ -494,10 +497,14 @@ EXPERIMENTAL: Interface subject to change."
                     ;; cycle counting isn't supported everywhere.
                     (when cycles
                       (note :processor-cycles cycles #'zerop))
-                    (note :lambdas-converted sb!c::*lambda-conversions* #'zerop)
+                    (note :lambdas-converted sb-c::*lambda-conversions* #'zerop)
                     (note :eval-calls *eval-calls* #'zerop)
-                    (note :gc-run-time-ms gc-internal-run-time)
+                    (note :gc-run-time-ms (floor gc-internal-run-time
+                                                 (/ internal-time-units-per-second 1000)))
+                    (note :gc-real-time-ms (floor gc-internal-real-time
+                                                  (/ internal-time-units-per-second 1000)))
                     (note :system-run-time-us system-run-time)
                     (note :user-run-time-us user-run-time)
-                    (note :real-time-ms real-time))
+                    (note :real-time-ms (floor real-time
+                                               (/ internal-time-units-per-second 1000))))
                   (apply timer plist))))))))))
