@@ -38,9 +38,7 @@
   ;; used for everything else. We use a separate EQ table to avoid
   ;; performance pathologies with objects for which SIMILAR
   ;; degenerates to EQL. Everything entered in the SIMILAR table is
-  ;; also entered in the EQ table. The NAMED-CONSTANT is used for
-  ;; named constants whose references are dumped as load time values
-  ;; of SYMBOL-GLOBAL-VALUE.
+  ;; also entered in the EQ table.
   (similar-table (make-similarity-table) :type hash-table :read-only t)
   (eq-table (make-hash-table :test 'eq) :type hash-table :read-only t)
   ;; the INSTANCE table maps dumpable instances to unique IDs for calculating
@@ -49,14 +47,6 @@
   ;; better not to add a new lazy stable hash slot to instances as a
   ;; side-effect of compiling.
   (instance-id-table (make-hash-table :test 'eq) :type hash-table :read-only t)
-  ;; the CONS table is an additional EQ table, used for storing CDRs
-  ;; of dumped lists which will not have their own direct identity as
-  ;; a dumped constant, but which might nevertheless be EQ to some
-  ;; other dumped object (and require that EQness to be preserved).
-  ;; The hash table entry, if present, is a reference to its parent
-  ;; list object (which will have a direct entity as a dumped
-  ;; constant) along with an index of how many CDRs to take.
-  (cons-table (make-hash-table :test 'eq) :type hash-table :read-only t)
   ;; Hashtable mapping a string to a list of fop-table indices of
   ;; symbols whose name is that string. For any name as compared
   ;; by STRING= there can be a symbol whose name is a base string
@@ -324,13 +314,24 @@
     (dump-fop 'fop-move-to-table fasl-output)
     (incf (fasl-output-table-free fasl-output))))
 
+(defun cdr-similarity-p (index fasl-output)
+  (when (consp index)
+    (destructuring-bind (list . nthcdr) index
+      (let ((index (gethash list (fasl-output-eq-table fasl-output))))
+        (when (fixnump index)
+          (dump-push index fasl-output)
+          (dump-fop 'fop-nthcdr fasl-output nthcdr)
+          t)))))
+
 ;;; If X is in File's SIMILAR-TABLE, then push the object and return T,
 ;;; otherwise NIL.
 (defun similar-check-table (x fasl-output)
   (declare (type fasl-output fasl-output))
-  (awhen (get-similar x (fasl-output-similar-table fasl-output))
-    (dump-push it fasl-output)
-    t))
+  (let ((index (get-similar x (fasl-output-similar-table fasl-output))))
+    (cond ((fixnump index)
+           (dump-push index fasl-output)
+           t)
+          ((cdr-similarity-p index fasl-output)))))
 
 ;;; These functions are called after dumping an object to save the
 ;;; object in the table. The object (also passed in as X) must already
@@ -446,8 +447,9 @@
 ;;; When we go to dump the object, we enter it in the CIRCULARITY-TABLE.
 (defun dump-non-immediate-object (x file)
   (let ((index (gethash x (fasl-output-eq-table file))))
-    (cond (index
+    (cond ((fixnump index)
            (dump-push index file))
+          ((cdr-similarity-p index file))
           (t
            (typecase x
              (symbol (dump-symbol x file))
@@ -458,8 +460,8 @@
                     ((not (similar-check-table x file))
                      (dump-list x file t)
                      (similar-save-object x file))))
-             (wrapper
-              (dump-wrapper x file)
+             (layout
+              (dump-layout x file)
               (eq-save-object x file))
              #+sb-xc-host
              (ctype
@@ -588,7 +590,6 @@
 ;;; We peek at the object type so that we only pay the circular
 ;;; detection overhead on types of objects that might be circular.
 (defun dump-object (x file)
-  #+(and metaspace sb-xc-host) (when (cl:typep x 'sb-vm:layout) (error "can't dump sb-vm:layout"))
   (if (compound-object-p x)
       (let ((*circularities-detected* ())
             (circ (fasl-output-circularity-table file)))
@@ -782,28 +783,25 @@
 
         ;; if this CONS is EQ to some other object we have already
         ;; dumped, dump a reference to that instead.
-        (let ((index (gethash l (fasl-output-eq-table file))))
-          (when index
-            (dump-push index file)
-            (terminate-dotted-list n file)
-            (return)))
+        (let* ((table (if coalesce
+                          (fasl-output-similar-table file)
+                          (fasl-output-eq-table file)))
+               (index (gethash l table)))
+          (cond ((fixnump index)
+                 (dump-push index file)
+                 (terminate-dotted-list n file)
+                 (return))
+                ((cdr-similarity-p index file)
+                 (when (> n 0)
+                   (terminate-dotted-list n file))
+                 (return)))
 
-        ;; if this CONS is EQ to the Ith CDR of some other list we have
-        ;; already dumped, dump a reference to that instead.
-        (let ((list+i (gethash l (fasl-output-cons-table file))))
-          (when list+i
-            (destructuring-bind (list i) list+i
-              (aver (consp list))
-              (let ((index (gethash list (fasl-output-eq-table file))))
-                (dump-push index file)
-                (dump-fop 'fop-nthcdr file i)
-                (when (> n 0)
-                  (terminate-dotted-list n file))
-                (return)))))
-
-        ;; put an entry for this cons into the fasl output cons table,
-        ;; for the benefit of dumping later constants
-        (setf (gethash l (fasl-output-cons-table file)) (list list n))
+          ;; put an entry for this cons into the fasl output cons table,
+          ;; for the benefit of dumping later constants
+          (let ((index (cons list n)))
+            (setf (gethash l (fasl-output-eq-table file)) index)
+            (when coalesce
+              (setf (gethash l (fasl-output-similar-table file)) index))))
 
         (setf (gethash l circ) list)
 
@@ -822,8 +820,9 @@
                  ;; This is the same as DUMP-NON-IMMEDIATE-OBJECT but
                  ;; without calling COALESCE-TREE-P again.
                  (let ((index (gethash obj (fasl-output-eq-table file))))
-                   (cond (index
+                   (cond ((fixnump index)
                           (dump-push index file))
+                         ((cdr-similarity-p index file))
                          ((not coalesce)
                           (dump-list obj file)
                           (eq-save-object obj file))
@@ -1032,7 +1031,7 @@
 
 (defconstant-eqx +fixup-flavors+
   #(:assembly-routine :assembly-routine*
-    :gc-barrier :symbol-tls-index
+    :card-table-index-mask :symbol-tls-index
     :alien-code-linkage-index :alien-data-linkage-index
     :foreign :foreign-dataref
     :code-object
@@ -1042,7 +1041,7 @@
   #'equalp)
 
 ;;; Pack the aspects of a fixup into an integer.
-;;; DATA is for asm routine fixups. The routine can be represented in 7 bits,
+;;; DATA is for asm routine fixups. The routine can be represented in 8 bits,
 ;;; so the fixup can be reduced to one word instead of an integer and a symbol.
 (declaim (inline !pack-fixup-info))
 (defun !pack-fixup-info (offset kind flavor data)
@@ -1053,18 +1052,18 @@
           (ash (the (mod 16) (or (position flavor +fixup-flavors+)
                                  (error "Bad fixup flavor ~s" flavor)))
                3)
-          ;; 7 bits
-          (ash (the (mod 128) data) 7)
+          ;; 8 bits
+          (ash (the (mod 256) data) 7)
           ;; whatever it needs
-          (ash offset 14)))
+          (ash offset 15)))
 
 ;;; Unpack an integer from DUMP-FIXUPs. Shared by genesis and target fasloader
 (declaim (inline !unpack-fixup-info))
 (defun !unpack-fixup-info (packed-info) ; Return (VALUES offset kind flavor data)
-  (values (ash packed-info -14)
+  (values (ash packed-info -15)
           (aref +fixup-kinds+ (ldb (byte 3 0) packed-info))
           (aref +fixup-flavors+ (ldb (byte 4 3) packed-info))
-          (ldb (byte 7 7) packed-info)))
+          (ldb (byte 8 7) packed-info)))
 
 #+(or arm arm64 riscv sparc) ; these don't have any retained packed fixups (yet)
 (defun sb-c::pack-retained-fixups (fixup-notes)
@@ -1086,29 +1085,25 @@
     (let* ((fixup (fixup-note-fixup note))
            (name (fixup-name fixup))
            (flavor (fixup-flavor fixup))
-           (named (not (member flavor '(:code-object :gc-barrier))))
+           (named (not (member flavor '(:code-object :card-table-index-mask))))
            (data
             (or #-sb-xc-host ; ASM routine indices aren't known to the cross-compiler
                 (when (member flavor '(:assembly-routine :assembly-routine*))
                   (setq named nil)
                   (the (integer 1 *) ; data can't be 0
-                       (cddr (gethash name (%code-debug-info *assembler-routines*)))))
+                       (cddr (gethash name (%asm-routine-table *assembler-routines*)))))
                 0))
            (info
             (!pack-fixup-info (fixup-note-position note) (fixup-note-kind note)
                               flavor data))
            (operand
             (ecase flavor
-              ((:code-object :gc-barrier) (the null name))
+              ((:code-object :card-table-index-mask) (the null name))
               (:layout
                (if (symbolp name)
                    name
-                   (wrapper-classoid-name
-                    (cond #+metaspace
-                          ((sb-kernel::layout-p name) (layout-friend name))
-                          (t name)))))
-              (:layout-id
-               (the wrapper name))
+                   (layout-classoid-name name)))
+              (:layout-id (the layout name))
               ((:assembly-routine :assembly-routine*
                :symbol-tls-index
                ;; Only #+immobile-space can use the following two flavors.
@@ -1192,7 +1187,11 @@
                 (dump-specialized-vector (make-array (cdr entry)
                                                      :element-type '(unsigned-byte 8)
                                                      :initial-element #xFF)
-                                         fasl-output))))
+                                         fasl-output))
+               #+arm64
+               (:tls-index
+                (dump-object (cdr entry) fasl-output)
+                (dump-fop 'fop-tls-index fasl-output))))
             (null
              (dump-fop 'fop-misc-trap fasl-output)))))
 
@@ -1239,6 +1238,9 @@
     ;; So these go on the stack last, i.e. nearest the top.
     ;; Reversing sorts the entry points in ascending address order
     ;; except possibly when there are multiple entry points to one routine
+    (unless (= (length (remove-duplicates (mapcar 'car routines)))
+               (length routines))
+      (error "Duplicated asm routine name"))
     (dolist (routine (reverse routines))
       (dump-object (car routine) file)
       (dump-integer (+ (label-position (cadr routine)) (caddr routine))
@@ -1391,12 +1393,12 @@
            struct))
   (note-potential-circularity struct file)
   (do* ((length (%instance-length struct))
-        (wrapper (%instance-wrapper struct))
-        (bitmap (wrapper-bitmap wrapper))
+        (layout (%instance-layout struct))
+        (bitmap (layout-bitmap layout))
         (circ (fasl-output-circularity-table file))
         (index sb-vm:instance-data-start (1+ index)))
       ((>= index length)
-       (dump-non-immediate-object wrapper file)
+       (dump-non-immediate-object layout file)
        (dump-fop 'fop-struct file length))
     (let* ((obj (if (logbitp index bitmap)
                     (%instance-ref struct index)
@@ -1413,14 +1415,14 @@
                              (t obj))
                        file))))
 
-(defun dump-wrapper (obj file &aux (flags (wrapper-flags obj)))
-  (when (wrapper-invalid obj)
+(defun dump-layout (obj file)
+  (when (layout-invalid obj)
     (compiler-error "attempt to dump reference to obsolete class: ~S"
-                    (wrapper-classoid obj)))
+                    (layout-classoid obj)))
   ;; STANDARD-OBJECT could in theory be dumpable, but nothing else,
   ;; because all its subclasses can evolve to have new layouts.
-  (aver (not (logtest flags +pcl-object-layout-flag+)))
-  (let ((name (wrapper-classoid-name obj)))
+  (aver (not (logtest (layout-flags obj) +pcl-object-layout-flag+)))
+  (let ((name (layout-classoid-name obj)))
     ;; Q: Shouldn't we aver that NAME is the proper name for its classoid?
     (unless name
       (compiler-error "dumping anonymous layout: ~S" obj))
@@ -1430,14 +1432,14 @@
     #-sb-xc-host
     (let ((fop (known-layout-fop name)))
       (when fop
-        (return-from dump-wrapper (dump-byte fop file))))
+        (return-from dump-layout (dump-byte fop file))))
     (dump-object name file))
-  (sub-dump-object (wrapper-bitmap obj) file)
-  (sub-dump-object (wrapper-inherits obj) file)
+  (sub-dump-object (layout-bitmap obj) file)
+  (sub-dump-object (layout-inherits obj) file)
   (dump-fop 'fop-layout file
-            (1+ (wrapper-depthoid obj)) ; non-stack args can't be negative
-            (logand flags sb-kernel::layout-flags-mask)
-            (wrapper-length obj)))
+            (1+ (layout-depthoid obj)) ; non-stack args can't be negative
+            (logand (layout-flags obj) sb-kernel::layout-flags-mask)
+            (layout-length obj)))
 
 ;;;; dumping instances which just save their slots
 
@@ -1472,3 +1474,21 @@
   (dump-object namestring file)
   (dump-object cc file)
   (dump-fop 'fop-record-code-coverage file))
+
+(defun dump-emitted-full-calls (hash-table fasl)
+  (let ((list (%hash-table-alist hash-table)))
+    #+sb-xc-host ; enforce host-insensitive reproducible ordering
+    (labels ((symbol< (a b)
+               (cond ((string< a b) t)
+                     ((string= a b)
+                      ;; this does find a few pairs of lookalikes
+                      (string< (cl:package-name (sb-xc:symbol-package a))
+                               (cl:package-name (sb-xc:symbol-package b))))))
+             (fname< (a b)
+               (cond ((and (atom a) (atom b)) (symbol< a b))
+                     ((atom a) t) ; symbol < list
+                     ((atom b) nil) ; opposite
+                     ((symbol< (cadr a) (cadr b))))))
+      (setq list (sort list #'fname< :key #'car)))
+    (dump-object list fasl)
+    (dump-fop 'fop-note-full-calls fasl)))
