@@ -98,12 +98,12 @@
 (defknown find-classoid (name-for-class &optional t)
   (or classoid null) ())
 (defknown classoid-of (t) classoid (flushable))
-(defknown wrapper-of (t) wrapper (flushable))
-(defknown wrapper-depthoid (wrapper) layout-depthoid (flushable))
-#+64-bit (defknown layout-depthoid (sb-vm:layout) layout-depthoid (flushable always-translatable))
-#+(or x86 x86-64) (defknown (layout-depthoid-ge) (sb-vm:layout integer) boolean (flushable))
+(defknown layout-of (t) layout (flushable))
+#+64-bit (defknown layout-depthoid (layout) layout-depthoid (flushable always-translatable))
+#+(or x86 x86-64) (defknown (layout-depthoid-ge) (layout integer) boolean (flushable))
 (defknown %structure-is-a (instance t) boolean (foldable flushable))
 (defknown structure-typep (t t) boolean (foldable flushable))
+(defknown classoid-cell-typep (t t) boolean (foldable flushable no-verify-arg-count))
 (defknown copy-structure (structure-object) structure-object
   (flushable)
   :derive-type #'result-type-first-arg)
@@ -432,7 +432,7 @@
 (defknown (float-digits float-precision) (float) float-digits
   (movable foldable unsafely-flushable))
 (defknown integer-decode-float (float)
-    (values integer float-int-exponent (member -1 1))
+    (values double-float-significand float-int-exponent (member -1 1))
     (movable foldable unsafely-flushable))
 
 (defknown single-float-sign (single-float) single-float (movable foldable flushable))
@@ -457,11 +457,12 @@
 (defknown lognot (integer) integer (movable foldable flushable))
 (defknown logtest (integer integer) boolean (movable foldable flushable commutative))
 (defknown logbitp (unsigned-byte integer) boolean (movable foldable flushable))
-(defknown ash (integer integer) integer
+(defknown (ash ash-inverted) (integer integer) integer
   (movable foldable flushable))
 (defknown %ash/right ((or word sb-vm:signed-word) (mod #.sb-vm:n-word-bits))
   (or word sb-vm:signed-word)
   (movable foldable flushable always-translatable))
+
 (defknown (logcount integer-length) (integer) bit-index
   (movable foldable flushable))
 ;;; FIXME: According to the ANSI spec, it's legal to use any
@@ -906,6 +907,8 @@
   (call important-result)
   :derive-type (creation-result-type-specifier-nth-arg 0))
 
+;; N.B., sb-simple-streams clobbers this; if this changes, make sure
+;; sb-simple-streams follows along, where necessary.
 (defknown read-sequence ((modifying sequence) stream
                          &key
                          (:start index)
@@ -1308,7 +1311,7 @@
    (:start (inhibit-flushing index 0))
    (:end (inhibit-flushing sequence-end nil)))
   simple-string (flushable)
-  :derive-type (sequence-result-nth-arg 0 :preserve-dimensions t))
+  :derive-type (sequence-result-nth-arg 0 :preserve-dimensions t :string-designator t))
 
 (defknown (nstring-upcase nstring-downcase nstring-capitalize)
   ((modifying string) &key (:start index) (:end sequence-end))
@@ -1477,8 +1480,12 @@
   ()
   :derive-type (read-elt-type-deriver t 'character nil))
 
+;; N.B., sb-simple-streams clobbers this; if this changes, make sure
+;; sb-simple-streams follows along, where necessary.
 (defknown listen (&optional stream-designator) boolean (flushable))
 
+;; N.B., sb-simple-streams clobbers this; if this changes, make sure
+;; sb-simple-streams follows along, where necessary.
 (defknown clear-input (&optional stream-designator) null ())
 
 (defknown read-from-string
@@ -1671,6 +1678,8 @@
 
 (defknown user-homedir-pathname (&optional t) pathname (flushable))
 
+;; N.B., sb-simple-streams clobbers this; if this changes, make sure
+;; sb-simple-streams follows along, where necessary.
 (defknown open
   (pathname-designator &key
                        (:class symbol)
@@ -1868,7 +1877,7 @@
 ;;; special optimizer anyway.
 (defknown %typep (t (or type-specifier ctype)) boolean
   (movable flushable no-verify-arg-count))
-(defknown %instance-typep (t (or type-specifier ctype wrapper)) boolean
+(defknown %instance-typep (t (or type-specifier ctype layout)) boolean
   (movable flushable always-translatable))
 ;;; We should never emit a call to %typep-wrapper
 (defknown %typep-wrapper (t t (or type-specifier ctype)) t
@@ -1906,7 +1915,7 @@
 (defknown %continue-unwind () nil)
 (defknown %throw (t &rest t) nil) ; This is MV-called.
 (defknown %nlx-entry (t) *)
-(defknown %%primitive (t t &rest t) *)
+(defknown %%primitive (t &rest t) *)
 (defknown %pop-values (t) t)
 (defknown %nip-values (t t &rest t) (values))
 (defknown %dummy-dx-alloc (t t) t)
@@ -1920,6 +1929,7 @@
 
 (defknown %odd-key-args-error () nil)
 (defknown %unknown-key-arg-error (t t) nil)
+(defknown nil-fun-returned-error (t) nil)
 (defknown (%ldb %mask-field) (bit-index bit-index integer) unsigned-byte
   (movable foldable flushable no-verify-arg-count))
 (defknown (%dpb %deposit-field) (integer bit-index bit-index integer) integer
@@ -1943,6 +1953,12 @@
 (defknown hairy-data-vector-set (array index t) t (no-verify-arg-count))
 (defknown hairy-data-vector-ref/check-bounds (array index) t (foldable no-verify-arg-count))
 (defknown hairy-data-vector-set/check-bounds (array index t) t (no-verify-arg-count))
+
+(defknown vector-hairy-data-vector-ref (vector index) t (foldable flushable no-verify-arg-count))
+(defknown vector-hairy-data-vector-set (vector index t) t (no-verify-arg-count))
+(defknown vector-hairy-data-vector-ref/check-bounds (vector index) t (foldable no-verify-arg-count))
+(defknown vector-hairy-data-vector-set/check-bounds (vector index t) t (no-verify-arg-count))
+
 (defknown %caller-frame () t (flushable))
 (defknown %caller-pc () system-area-pointer (flushable))
 (defknown %with-array-data (array index (or index null))
@@ -2098,7 +2114,7 @@
 (defknown %program-error (&optional t &rest t) nil ())
 (defknown compiler-error (t &rest t) nil ())
 (defknown (compiler-warn compiler-style-warn) (t &rest t) (values) ())
-(defknown (compiler-mumble note-lossage note-unwinnage) (string &rest t) (values) ())
+(defknown (compiler-mumble note-lossage note-unwinnage) (format-control &rest t) (values) ())
 (defknown (compiler-notify maybe-compiler-notify) ((or format-control symbol) &rest t)
   (values)
   ())
@@ -2230,22 +2246,13 @@
   (defknown round-single (single-float #1#) single-float
       (foldable flushable movable always-translatable)))
 
-(defknown fixnum*
-  (fixnum fixnum t)
-  fixnum
-  (movable foldable always-translatable))
+(defknown (overflow* overflow+ overflow-
+           overflow-ash)
+  (integer integer t)
+  integer
+  (movable always-translatable))
 
-(defknown (signed* signed+ signed-)
-  (sb-vm:signed-word sb-vm:signed-word t)
-  sb-vm:signed-word
-  (movable foldable commutative always-translatable))
-
-(defknown (unsigned* unsigned+ unsigned-)
-  (word word t)
-  word
-  (movable foldable commutative always-translatable))
-
-(defknown (unsigned+signed unsigned-signed)
-  (word sb-vm:signed-word t)
-  word
-  (movable foldable commutative always-translatable))
+(defknown overflow-negate
+  (integer t)
+  integer
+  (movable always-translatable))
