@@ -43,20 +43,25 @@
 
 ;; Numerics including fixnum, excluding short-float. (INTEGER,RATIONAL)
 (defun %test-fixnum-and-headers (value temp target not-p headers
-                                 &key value-tn-ref)
+                                 &key value-tn-ref immediate-tested)
   (let ((drop-through (gen-label)))
     (case n-fixnum-tag-bits
-     (1 (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+     (1
+      (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+      (when (types-equal-or-intersect (tn-ref-type value-tn-ref)
+                                      (specifier-type 'fixnum))
         (inst test :byte temp 1)
-        (inst jmp :nz (if not-p drop-through target)) ; inverted
-        (%test-headers value temp target not-p nil headers
-                       :drop-through drop-through :compute-temp nil
-                       :value-tn-ref value-tn-ref))
+        (inst jmp :nz (if not-p drop-through target))) ; inverted
+      (%test-headers value temp target not-p nil headers
+                     :drop-through drop-through :compute-temp nil
+                     :value-tn-ref value-tn-ref
+                     :immediate-tested immediate-tested))
      (t
       (generate-fixnum-test value)
       (inst jmp :z (if not-p drop-through target))
       (%test-headers value temp target not-p nil headers
-                     :drop-through drop-through)))))
+                     :drop-through drop-through
+                     :immediate-tested immediate-tested)))))
 
 ;; I can see no reason this would ever be used.
 ;; (or fixnum character|unbound-marker) is implausible.
@@ -69,21 +74,30 @@
 
 ;; Numerics
 (defun %test-fixnum-immediate-and-headers (value temp target not-p immediate headers
-                                           &key value-tn-ref)
+                                           &key value-tn-ref immediate-tested)
   (let ((drop-through (gen-label)))
     (case n-fixnum-tag-bits
-     (1 (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+     (1
+      (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+      (when (types-equal-or-intersect (tn-ref-type value-tn-ref)
+                                      (specifier-type 'fixnum))
         (inst test :byte temp 1)
-        (inst jmp :nz (if not-p drop-through target)) ; inverted
+        (inst jmp :nz (if not-p drop-through target))) ; inverted
+      (when (or (/= immediate single-float-widetag)
+                (types-equal-or-intersect (tn-ref-type value-tn-ref)
+                                          (specifier-type 'single-float)))
         (inst cmp :byte temp (- immediate other-pointer-lowtag))
-        (inst jmp :e (if not-p drop-through target))
-        (%test-headers value temp target not-p nil headers
-                       :drop-through drop-through :compute-temp nil
-                       :value-tn-ref value-tn-ref))
+        (inst jmp :e (if not-p drop-through target)))
+
+      (%test-headers value temp target not-p nil headers
+                     :drop-through drop-through :compute-temp nil
+                     :value-tn-ref value-tn-ref
+                     :immediate-tested immediate-tested))
      (t (generate-fixnum-test value)
         (inst jmp :z (if not-p drop-through target))
         (%test-immediate-and-headers value temp target not-p immediate headers
-                                     :drop-through drop-through)))))
+                                     :drop-through drop-through
+                                     :immediate-tested immediate-tested)))))
 
 (defun %test-immediate (value temp target not-p immediate
                         &optional (drop-through (gen-label)))
@@ -95,7 +109,8 @@
 ;; Numerics including short-float, excluding fixnum
 (defun %test-immediate-and-headers (value temp target not-p immediate headers
                                     &key (drop-through (gen-label))
-                                         value-tn-ref)
+                                         value-tn-ref
+                                         immediate-tested)
   ;; Code a single instruction byte test if possible.
   (cond ((sc-is value any-reg descriptor-reg)
          (inst cmp :byte value immediate))
@@ -105,7 +120,8 @@
   (inst jmp :e (if not-p drop-through target))
   (%test-headers value temp target not-p nil headers
                  :drop-through drop-through
-                 :value-tn-ref value-tn-ref))
+                 :value-tn-ref value-tn-ref
+                 :immediate-tested immediate-tested))
 
 (defun %test-lowtag (value temp target not-p lowtag)
   (%lea-for-lowtag-test temp value lowtag)
@@ -117,7 +133,8 @@
                            (drop-through (gen-label))
                            (load-widetag t)
                            (compute-temp load-widetag)
-                           value-tn-ref)
+                           value-tn-ref
+                           immediate-tested)
   (let* ((lowtag (if function-p fun-pointer-lowtag other-pointer-lowtag))
          ;; It is preferable (smaller and faster code) to directly
          ;; compare the value in memory instead of loading it into
@@ -149,7 +166,7 @@
       (cond ((not load-widetag))
             ((and value-tn-ref
                   (eq lowtag other-pointer-lowtag)
-                  (other-pointer-tn-ref-p value-tn-ref t))) ; best case: lowtag is right
+                  (other-pointer-tn-ref-p value-tn-ref t immediate-tested))) ; best case: lowtag is right
             ((and value-tn-ref
                   ;; If HEADERS contains a range, then list pointers have to be
                   ;; disallowed - consider a list whose CAR has a fixnum that
@@ -238,7 +255,7 @@
 (define-vop (simple-type-predicate)
   (:args (value :scs (any-reg descriptor-reg control-stack)))
   (:conditional)
-  (:args-var args)
+  (:arg-refs args)
   (:policy :fast-safe))
 
 (define-vop (fixnump/unsigned-byte-64 simple-type-predicate)
@@ -296,7 +313,7 @@
 (define-vop (signed-byte-64-p pointerp)
   (:translate signed-byte-64-p)
   (:conditional :z)
-  (:args-var arg-ref)
+  (:arg-refs arg-ref)
   (:generator 6
     (when (types-equal-or-intersect (tn-ref-type arg-ref) (specifier-type 'fixnum))
       (inst test :byte value fixnum-tag-mask)
@@ -326,7 +343,7 @@
                   (:translate ,name)
                   (:args (value :scs (any-reg descriptor-reg)))
                   (:conditional :z)
-                  (:args-var arg-ref)
+                  (:arg-refs arg-ref)
                   (:policy :fast-safe)
                   (:temporary (:sc unsigned-reg) temp temp2)
                   (:generator 6
@@ -366,7 +383,11 @@
 ;; the header of which must not overlap the static alloc regions
 #-sb-thread
 (aver (>= (- non-negative-fixnum-mask-constant-wired-address (* 2 n-word-bytes))
-          (+ (max boxed-region cons-region mixed-region) (* 3 n-word-bytes))))
+          (+ static-space-start
+             (max boxed-region-offset
+                  cons-region-offset
+                  mixed-region-offset)
+             (* 3 n-word-bytes))))
 
 ;;; An (unsigned-byte 64) can be represented with either a positive
 ;;; fixnum, a bignum with exactly one positive digit, or a bignum with
@@ -374,19 +395,26 @@
 (define-vop (unsigned-byte-64-p type-predicate)
   (:translate unsigned-byte-64-p)
   (:generator 10
-    (let ((not-target (gen-label))
-          (single-word (gen-label))
-          (fixnum-p (types-equal-or-intersect (tn-ref-type args) (specifier-type 'fixnum))))
+    (let* ((not-target (gen-label))
+           (single-word (gen-label))
+           (fixnum-p (types-equal-or-intersect (tn-ref-type args) (specifier-type 'fixnum)))
+           (not-signed-byte-64-p (not (types-equal-or-intersect (tn-ref-type args) (specifier-type 'signed-word))))
+           (unsigned-p (or not-signed-byte-64-p
+                           (not (types-equal-or-intersect (tn-ref-type args) (specifier-type '(integer * -1)))))))
       (multiple-value-bind (yep nope)
           (if not-p
               (values not-target target)
               (values target not-target))
         (when fixnum-p
-          ;; Is it a fixnum with the sign bit clear?
-          (inst test (ea non-negative-fixnum-mask-constant-wired-address) value)
-          (inst jmp :z yep))
+          (cond (unsigned-p
+                 (inst test :byte value fixnum-tag-mask)
+                 (inst jmp :z yep))
+                (t ;; Is it a fixnum with the sign bit clear?
+                 (inst test (ea non-negative-fixnum-mask-constant-wired-address) value)
+                 (inst jmp :z yep))))
         (cond ((fixnum-or-other-pointer-tn-ref-p args t)
-               (when fixnum-p
+               (when (and fixnum-p
+                          (not unsigned-p))
                  (inst test :byte value fixnum-tag-mask)
                  (inst jmp :z nope)))
               (t
@@ -395,9 +423,12 @@
                (inst jmp :ne nope)))
         ;; Get the header.
         (loadw temp value 0 other-pointer-lowtag)
-        ;; Is it one?
-        (inst cmp temp (bignum-header-for-length 1))
-        (inst jmp :e single-word)
+        (unless not-signed-byte-64-p
+          ;; Is it one?
+          (inst cmp temp (bignum-header-for-length 1))
+          (inst jmp :e (if unsigned-p
+                           yep
+                           single-word)))
         ;; If it's other than two, we can't be an (unsigned-byte 64)
         ;: Leave TEMP holding 0 in the affirmative case.
         (inst sub temp (bignum-header-for-length 2))
@@ -405,15 +436,19 @@
         ;; Compare the second digit to zero (in TEMP).
         (inst cmp (object-slot-ea value (1+ bignum-digits-offset) other-pointer-lowtag)
               temp)
-        (inst jmp :z yep) ; All zeros, its an (unsigned-byte 64).
-        (inst jmp nope)
+        (cond (unsigned-p
+               (inst jmp (if not-p :nz :z) target))
+              (t
+               (inst jmp :z yep) ; All zeros, its an (unsigned-byte 64).
+               (inst jmp nope)))
 
-        (emit-label single-word)
-        ;; Get the single digit.
-        (loadw temp value bignum-digits-offset other-pointer-lowtag)
-        ;; positive implies (unsigned-byte 64).
-        (inst test temp temp)
-        (inst jmp (if not-p :s :ns) target)
+        (unless unsigned-p
+          (emit-label single-word)
+          ;; Get the single digit.
+          (loadw temp value bignum-digits-offset other-pointer-lowtag)
+          ;; positive implies (unsigned-byte 64).
+          (inst test temp temp)
+          (inst jmp (if not-p :s :ns) target))
 
         (emit-label not-target)))))
 
@@ -430,16 +465,22 @@
   (:conditional :e)
   (:info hi)
   (:policy :fast-safe)
+  (:arg-refs value-ref)
+  (:vop-var vop)
   (:generator 4
-     (let* ((fixnum-hi (if (sc-is value unsigned-reg signed-reg)
-                           hi
-                           (fixnumize hi)))
-            (mask (lognot fixnum-hi))
-            (constant
-             (if (= (ldb (byte 64 0) mask) non-negative-fixnum-mask-constant)
-                 (ea non-negative-fixnum-mask-constant-wired-address)
-                 (constantize mask))))
-       (inst test value constant))))
+    (let ((fixnum-hi (if (sc-is value unsigned-reg signed-reg)
+                         hi
+                         (fixnumize hi))))
+      (cond ((sc-is value descriptor-reg)
+             (let* ((mask (lognot fixnum-hi))
+                    (constant
+                      (if (= (ldb (byte 64 0) mask) non-negative-fixnum-mask-constant)
+                          (ea non-negative-fixnum-mask-constant-wired-address)
+                          (constantize mask))))
+               (inst test value constant)))
+            (t
+             (change-vop-flags vop '(:be))
+             (inst cmp value (constantize fixnum-hi)))))))
 
 (define-vop (test-fixnum-mod-tagged-unsigned)
   (:args (value :scs (any-reg unsigned-reg signed-reg)))
@@ -552,7 +593,7 @@
                   (:translate ,name)
                   (:info) ; nullify the info
                   (:conditional :z)
-                  (:args-var value-tn-ref)
+                  (:arg-refs value-tn-ref)
                   (:generator 4
                     (fail-if-not-otherptr)
                     (inst or :byte temp 8)
@@ -569,7 +610,7 @@
                     (:translate ,name)
                     (:info)
                     (:conditional :c) ; Carry flag = "below" (unsigned)
-                    (:args-var value-tn-ref)
+                    (:arg-refs value-tn-ref)
                     (:generator 4
                       (fail-if-not-otherptr)
                       (inst sub :byte temp ,min)
@@ -645,9 +686,8 @@
 (eval-when (:compile-toplevel) (aver (= sb-impl::package-id-bits 16)))
 (define-vop (keywordp symbolp)
   (:translate keywordp)
-  (:args-var args-ref)
   (:generator 3
-    (cond ((csubtypep (tn-ref-type args-ref) (specifier-type 'symbol))
+    (cond ((csubtypep (tn-ref-type args) (specifier-type 'symbol))
            (inst cmp :word (ea (+ (ash symbol-name-slot word-shift) 6
                                   (- other-pointer-lowtag))
                                value)
@@ -678,7 +718,7 @@
   (:info widetag)
   (:arg-types * (:constant t))
   (:conditional :e)
-  (:args-var args)
+  (:arg-refs args)
   (:generator 2
    (inst cmp :byte (ea (- other-pointer-lowtag) x) widetag)))
 
@@ -703,7 +743,7 @@
    (:node-var node)
    (:generator 1
      (unless (sb-c::set-slot-old-p node)
-       (emit-gc-store-barrier object nil temp (vop-nth-arg 1 vop) value))
+       (emit-gengc-barrier object nil temp (vop-nth-arg 1 vop) value))
      (inst mov :dword (ea (- 4 instance-pointer-lowtag) object) value)))
  (define-vop (%fun-layout %instance-layout)
    (:translate %fun-layout)
@@ -712,22 +752,19 @@
    (:translate %set-fun-layout)
    (:generator 1
      (unless (sb-c::set-slot-old-p node)
-       (emit-gc-store-barrier object nil temp (vop-nth-arg 1 vop) value))
+       (emit-gengc-barrier object nil temp (vop-nth-arg 1 vop) value))
      (inst mov :dword (ea (- 4 fun-pointer-lowtag) object) value)))
  (define-vop ()
   (:translate sb-c::layout-eq)
   (:policy :fast-safe)
   (:conditional :e)
   (:args (object :scs (descriptor-reg))
-         (layout :scs (descriptor-reg immediate #+metaspace constant)))
+         (layout :scs (descriptor-reg immediate)))
   (:arg-types * * (:constant t))
   (:info lowtag)
   (:generator 1
-    ;; With metaspace, the layout argument is actually a #<WRAPPER>
-    ;; which does not have IMMEDIATE sc, but rather CONSTANT sc.
-    ;; But we use a layout fixup which stuffs in the pointer to the layout.
     (inst cmp :dword (ea (- 4 lowtag) object)
-          (if (sc-is layout immediate constant)
+          (if (sc-is layout immediate)
               (make-fixup (tn-value layout) :layout)
               layout)))))
 
@@ -759,7 +796,7 @@
 
 (define-vop (fixnump simple-type-predicate)
   (:translate fixnump)
-  (:args-var arg-ref)
+  (:arg-refs arg-ref)
   (:args (value :scs (any-reg descriptor-reg) :load-if (tn-ref-memory-access arg-ref)))
   (:conditional :z)
   ;; the compiler is very sensitive to this cost here as regards boxing. DON'T TOUCH !!!
@@ -829,18 +866,29 @@
   (:temporary (:sc unsigned-reg) temp)
   (:conditional)
   (:info target not-p)
-  (:args-var args)
+  (:arg-refs integer-ref)
   (:policy :fast-safe)
   (:variant-vars comparison)
   (:variant :g)
   (:generator 8
-    (unless (sc-is (tn-ref-tn args) descriptor-reg control-stack)
-      (setf args (tn-ref-across args)))
-    (let* ((integer-p (csubtypep (tn-ref-type args) (specifier-type 'integer)))
+    (let* ((integer-p (csubtypep (tn-ref-type integer-ref) (specifier-type 'integer)))
            (fixnum (if (sc-is fixnum immediate)
-                       (let* ((value (fixnumize (tn-value fixnum))))
+                       (let* ((value (fixnumize (tn-value fixnum)))
+                              (one (fixnumize 1)))
                          (cond ((plausible-signed-imm32-operand-p value)
                                 value)
+                               ((and (plausible-signed-imm32-operand-p (+ value one))
+                                     (or (and (eql comparison :le)
+                                              (setf comparison :l))
+                                         (and (eql comparison :g)
+                                              (setf comparison :ge))))
+                                (setf value (+ value one)))
+                               ((and (plausible-signed-imm32-operand-p (- value one))
+                                     (or (and (eql comparison :ge)
+                                              (setf comparison :g))
+                                         (and (eql comparison :l)
+                                              (setf comparison :le))))
+                                (setf value (- value one)))
                                (t
                                 (inst mov temp value)
                                 temp)))
@@ -850,16 +898,28 @@
               (values not-target target)
               (values target not-target))
         (assemble ()
-          (when (types-equal-or-intersect (tn-ref-type args) (specifier-type 'fixnum))
+          (when (types-equal-or-intersect (tn-ref-type integer-ref) (specifier-type 'fixnum))
             (generate-fixnum-test integer)
             (inst jmp :nz BIGNUM)
-            (if (eql fixnum 0)
-                (inst test integer integer)
-                (inst cmp integer fixnum))
+            (cond ((eql fixnum 0)
+                   (inst test integer integer))
+                  ((or (and (eql fixnum #.(fixnumize -1))
+                            (or (and (eql comparison :le)
+                                     (setf comparison :l))
+                                (and (eql comparison :g)
+                                     (setf comparison :ge))))
+                       (and (eql fixnum #.(fixnumize 1))
+                            (or (and (eql comparison :ge)
+                                     (setf comparison :g))
+                                (and (eql comparison :l)
+                                     (setf comparison :le)))))
+                   (inst test integer integer))
+                  (t
+                   (inst cmp integer fixnum)))
             (inst jmp comparison yep)
             (inst jmp nope))
           bignum
-          (unless (fixnum-or-other-pointer-tn-ref-p args t)
+          (unless (fixnum-or-other-pointer-tn-ref-p integer-ref t)
             (test-type integer temp nope t (other-pointer-lowtag)))
           (loadw temp integer 0 other-pointer-lowtag)
           (unless integer-p
@@ -883,6 +943,7 @@
   (:args (fixnum :scs (immediate any-reg))
          (integer :scs (descriptor-reg)))
   (:arg-types tagged-num (:or integer bignum))
+  (:arg-refs nil integer-ref)
   (:variant :l))
 
 (define-vop (<-fixnum-integer >-fixnum-integer)
@@ -905,16 +966,15 @@
 
 (define-vop (load-other-pointer-widetag)
   (:args (value :scs (any-reg descriptor-reg)))
-  (:args-var args)
-  (:info not-other-pointer-label null-label)
+  (:arg-refs args)
+  (:info not-other-pointer-label null-label zero-extend)
   (:results (r :scs (unsigned-reg)))
   (:result-types unsigned-num)
   (:generator 1
-    ;; FIXME: can we pass T as the 2nd arg to OTHER-POINTER-TN-REF-P here?
-    ;; I'm confused as to the intent of NULL-LABEL.
-    ;; Must branching occur if nil, or _may_ branching occur?
-    (cond ((other-pointer-tn-ref-p args)
-           (inst mov :byte r (ea (- other-pointer-lowtag) value)))
+    (cond ((other-pointer-tn-ref-p args (not null-label))
+           (if zero-extend
+               (inst movzx '(:byte :dword) r (ea (- other-pointer-lowtag) value))
+               (inst mov :byte r (ea (- other-pointer-lowtag) value))))
           (t
            (when null-label
              (inst cmp value nil-value)
@@ -922,7 +982,9 @@
            (%lea-for-lowtag-test r value other-pointer-lowtag :qword)
            (inst test :byte r lowtag-mask)
            (inst jmp :nz not-other-pointer-label)
-           (inst mov :byte r (ea r))))))
+           (if zero-extend
+               (inst movzx '(:byte :dword) r (ea r))
+               (inst mov :byte r (ea r)))))))
 
 (define-vop (test-widetag)
   (:args (value :scs (unsigned-reg) :target temp))
@@ -930,7 +992,10 @@
   (:info target not-p type-codes)
   (:generator 1
     (move temp value :dword)
-    (%test-headers nil temp target not-p nil type-codes
+    (%test-headers nil temp target not-p nil
+      (if (every #'integerp type-codes)
+          (canonicalize-widetags type-codes)
+          type-codes)
       :load-widetag nil)))
 
 (macrolet ((read-depthoid ()
@@ -969,7 +1034,7 @@
                         instance-pointer-lowtag)
                      layout)
                  test-layout))
-          ((let ((classoid (wrapper-classoid test-layout)))
+          ((let ((classoid (layout-classoid test-layout)))
              (and (eq (classoid-state classoid) :sealed)
                   (not (classoid-subclasses classoid))))
            (emit-constant test-layout)
@@ -980,7 +1045,7 @@
            (inst cmp (emit-constant test-layout) layout))
 
           (t
-           (let* ((depthoid (wrapper-depthoid test-layout))
+           (let* ((depthoid (layout-depthoid test-layout))
                   (offset (+ (id-bits-offset)
                              (ash (- depthoid 2) 2)
                              (- instance-pointer-lowtag))))
@@ -1016,7 +1081,7 @@
   (:translate sb-c::structure-typep)
   (:args (object :scs (descriptor-reg)))
   (:arg-types * (:constant t))
-  (:args-var args)
+  (:arg-refs args)
   (:policy :fast-safe)
   (:conditional)
   (:info target not-p test-layout)
@@ -1026,7 +1091,7 @@
       (%test-lowtag object layout (if not-p target done) t instance-pointer-lowtag))
 
     (cond ((and (not (integerp test-layout))
-                (let ((classoid (wrapper-classoid test-layout)))
+                (let ((classoid (layout-classoid test-layout)))
                   (and (eq (classoid-state classoid) :sealed)
                        (not (classoid-subclasses classoid)))))
            (emit-constant test-layout)
@@ -1054,7 +1119,6 @@
   (:args (layout :scs (descriptor-reg)))
   (:arg-types * (:constant t))
   (:policy :fast-safe)
-  (:conditional)
   (:info target not-p test-layout)
   (:generator 4
     (structure-is-a layout test-layout target not-p done)
@@ -1066,7 +1130,7 @@
 
 (define-vop (load-instance-layout)
   (:args (object :scs (any-reg descriptor-reg)))
-  (:args-var args)
+  (:arg-refs args)
   (:info not-instance)
   (:temporary (:sc unsigned-reg) temp)
   (:results (r :scs (descriptor-reg)))
