@@ -41,6 +41,7 @@
 #include "gc-internal.h"
 #include "gc-private.h"
 #include "code.h"
+#include "graphvisit.h"
 
 #include <errno.h>
 
@@ -218,12 +219,11 @@ static void inflate_core_bytes(int fd, os_vm_offset_t offset,
 }
 #endif
 
-#define MAX_SPACE_RELOCATION_RANGES 4
 struct heap_adjust {
     struct range {
         lispobj start, end;
         sword_t delta;
-    } range[MAX_SPACE_RELOCATION_RANGES];
+    } range[MAX_CORE_SPACE_ID+1];
     int n_ranges;
     int n_relocs_abs; // absolute
     int n_relocs_rel; // relative
@@ -270,7 +270,6 @@ set_adjustment(struct heap_adjust* adj,
     sword_t delta = len ? actual_addr - desired_addr : 0;
     if (!delta) return;
     int j = adj->n_ranges;
-    gc_assert(j < MAX_SPACE_RELOCATION_RANGES);
     adj->range[j].start = (lispobj)desired_addr;
     adj->range[j].end   = (lispobj)desired_addr + len;
     adj->range[j].delta = delta;
@@ -377,6 +376,26 @@ static inline void fix_fun_header_layout(lispobj __attribute__((unused)) *fun,
 #endif
 }
 
+/* Refer to the figures above CALCULATE-DD-BITMAP in src/code/defstruct
+ * for the meaning of these constants.
+ * I think this fixes what may have been a latent bug in space relocation.
+ * Due to inconsistent use of the bitmap, it could have processed funinstance
+ * LAYOUT slots twice.  Layout is supposed to have a 0 bit, meaning do NOT scan
+ * in a loop over slots, but instead externally to the loop. By marking it as 1,
+ * it could be processed both outside and inside the loop. In the extremely rare
+ * case of partial overlap of the desired and actual heap addresses, a pointer
+ * somewhere in the middle can not be discerned as having been fixed up or not
+ * fixed up based on its value. You have to know based on code flow.
+ * Therefore layouts slots always have to be treated as raw in the bitmap,
+ * but I did not see where it is enforced in lisp for funinstances. */
+#ifndef LISP_FEATURE_EXECUTABLE_FUNINSTANCES
+# define FUNINSTANCE_BITMAP -4
+#elif defined LISP_FEATURE_COMPACT_INSTANCE_HEADER
+# define FUNINSTANCE_BITMAP -8
+#else
+# define FUNINSTANCE_BITMAP -24
+#endif
+
 static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
 {
     lispobj *where = (lispobj*)start;
@@ -398,14 +417,6 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
         widetag = header_widetag(word);
         nwords = sizetab[widetag](where);
         switch (widetag) {
-        case FUNCALLABLE_INSTANCE_WIDETAG:
-            // Special note on the word at where[1] in funcallable instances:
-            // - If no immobile code, then the word points to read-only space,
-            ///  hence needs no adjustment.
-            // - Otherwise, the word might point to a relocated range,
-            //   either the instance itself, or a trampoline in immobile space.
-            adjust_word_at(where+1, adj);
-            /* FALLTHROUGH */
         case INSTANCE_WIDETAG:
             layout = layout_of(where);
             adjusted_layout = adjust_word(adj, layout);
@@ -413,11 +424,32 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
             // so it would not be fixed up otherwise.
             if (adjusted_layout != layout) layout_of(where) = adjusted_layout;
             struct bitmap bitmap = get_layout_bitmap(LAYOUT(adjusted_layout));
+            {
             lispobj* slots = where+1;
-            for (i=0; i<(nwords-1); ++i)
+            for (i=0; i<(nwords-1); ++i) // -1 from nwords because 'slots' is +1 from 'where'
                 if (bitmap_logbitp(i, bitmap)) adjust_pointers(slots+i, 1, adj);
+            }
             continue;
-          case SYMBOL_WIDETAG:
+        case FUNCALLABLE_INSTANCE_WIDETAG:
+            /* If non-executable funinstance, then word index 1 points at read-only space,
+             * hence needs no adjustment. Otherwise, the word points within the funinstance.
+             * Either way, adjust_word_at will do the right thing */
+            adjust_word_at(where+1, adj);
+            /* We have to adjust the layout, but its bitmap can be disregarded.
+             * There is no variation in bitmap by particular subtype of funinstance */
+            layout = funinstance_layout(where);
+            adjusted_layout = adjust_word(adj, layout);
+            if (adjusted_layout != layout) funinstance_layout(where) = adjusted_layout;
+            const int mask = FUNINSTANCE_BITMAP;
+            gc_assert((mask & (1<<0)) == 0); // bitmap must reflect it as untagged
+            {
+            lispobj* slots = where+1;
+            // slots[0] is where[1] which we just adjusted, so skip it.
+            for (i=1; i<(nwords-1); ++i) // -1 from nwords because 'slots' is +1 from 'where'
+                if ((1<<i) & mask) adjust_pointers(slots+i, 1, adj);
+            }
+            continue;
+        case SYMBOL_WIDETAG:
             { // Copied from scav_symbol() in gc-common
             struct symbol* s = (void*)where;
             adjust_pointers(&s->value, 3, adj); // value, fdefn, info
@@ -493,7 +525,7 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
               lispobj* data = (lispobj*)v->data;
               adjust_pointers(&data[vector_len(v)-1], 1, adj);
               int hwm = KV_PAIRS_HIGH_WATER_MARK(data);
-              boolean needs_rehash = 0;
+              bool needs_rehash = 0;
               lispobj *where = &data[2], *end = &data[2*(hwm+1)];
               // Adjust the elements, checking for need to rehash.
               for ( ; where < end ; where += 2) {
@@ -539,24 +571,8 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
                 FIXUP(where[1] += delta, where+1);
             }
             continue;
-        case BIGNUM_WIDETAG:
-#ifndef LISP_FEATURE_64_BIT
-        case SINGLE_FLOAT_WIDETAG:
-#endif
-        case DOUBLE_FLOAT_WIDETAG:
-        case COMPLEX_SINGLE_FLOAT_WIDETAG:
-        case COMPLEX_DOUBLE_FLOAT_WIDETAG:
-#ifdef SIMD_PACK_WIDETAG
-        case SIMD_PACK_WIDETAG:
-#endif
-#ifdef SIMD_PACK_256_WIDETAG
-        case SIMD_PACK_256_WIDETAG:
-#endif
-        case FILLER_WIDETAG: // non-card-spanning object pages container fillers
-            continue;
         default:
-          if (other_immediate_lowtag_p(widetag)
-              && specialized_vector_widetag_p(widetag))
+          if (other_immediate_lowtag_p(widetag) && leaf_obj_widetag_p(widetag))
               continue;
           else
               lose("Unrecognized heap object: @%p: %"OBJ_FMTX, where, *where);
@@ -581,6 +597,12 @@ static void relocate_heap(struct heap_adjust* adj)
                         (char*)adj->range[i].start + adj->range[i].delta,
                         (char*)adj->range[i].end + adj->range[i].delta);
     }
+#ifdef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
+    relocate_space(READ_ONLY_SPACE_START, read_only_space_free_pointer, adj);
+    // Relocate the CAR slot of nil-as-a-list, which needs to point to
+    // itself.
+    adjust_pointers((void*)(NIL - LIST_POINTER_LOWTAG), 1, adj);
+#endif
     relocate_space(NIL_SYMBOL_SLOTS_START, (lispobj*)NIL_SYMBOL_SLOTS_END, adj);
     relocate_space(STATIC_SPACE_OBJECTS_START, static_space_free_pointer, adj);
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
@@ -602,12 +624,7 @@ static void relocate_heap(struct heap_adjust* adj)
 /// Compute the bounds of the lisp assembly routine code object
 void calc_asm_routine_bounds()
 {
-#ifdef LISP_FEATURE_METASPACE
-    if (widetag_of((lispobj*)READ_ONLY_SPACE_START) == CODE_HEADER_WIDETAG)
-        asm_routines_start = READ_ONLY_SPACE_START;
-    else
-        asm_routines_start = READ_ONLY_SPACE_START + (256+2)*N_WORD_BYTES;
-#elif defined LISP_FEATURE_IMMOBILE_CODE
+#if defined LISP_FEATURE_IMMOBILE_CODE
     asm_routines_start = TEXT_SPACE_START;
 #else
     if ((uword_t)read_only_space_free_pointer > READ_ONLY_SPACE_START &&
@@ -755,7 +772,7 @@ process_directory(int count, struct ndir_entry *entry,
         extern int alien_linkage_table_n_prelinked;
         count = alien_linkage_table_n_prelinked = *ptr++;
         for ( ; count-- ; entry_index++ ) {
-            boolean datap = *ptr == (lispobj)-1; // -1 can't be a function address
+            bool datap = *ptr == (lispobj)-1; // -1 can't be a function address
             if (datap)
                 ++ptr;
             arch_write_linkage_table_entry(entry_index, (void*)*ptr++, datap);
@@ -768,7 +785,7 @@ process_directory(int count, struct ndir_entry *entry,
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     {
         spaces[IMMOBILE_FIXEDOBJ_CORE_SPACE_ID].desired_size +=
-            TEXT_SPACE_SIZE + ALIEN_LINKAGE_TABLE_SPACE_SIZE;
+            text_space_size + ALIEN_LINKAGE_TABLE_SPACE_SIZE;
     }
 #endif
 
@@ -786,9 +803,12 @@ process_directory(int count, struct ndir_entry *entry,
             lose("unknown space ID %ld addr %p", id, (void*)addr);
 
 #ifdef LISP_FEATURE_DARWIN_JIT
-        int enforce_address = (id == STATIC_CORE_SPACE_ID) || (id == READ_ONLY_CORE_SPACE_ID);
+        int enforce_address = (id == READ_ONLY_CORE_SPACE_ID);
 #else
-        int enforce_address = id == STATIC_CORE_SPACE_ID;
+        int enforce_address = 0;
+#endif
+#ifndef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
+        enforce_address |= (id == STATIC_CORE_SPACE_ID);
 #endif
 
         // We'd like to enforce proper alignment of 'addr' but there's
@@ -807,22 +827,38 @@ process_directory(int count, struct ndir_entry *entry,
                  (unsigned long)dynamic_space_size >> 10);
         }
 #ifndef LISP_FEATURE_DARWIN_JIT
+#ifdef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
+        if (id == READ_ONLY_CORE_SPACE_ID || id == STATIC_CORE_SPACE_ID) {
+#else
         if (id == READ_ONLY_CORE_SPACE_ID) {
-            if (len) // There is no "nominal" size of readonly space, so give it a size
+#endif
+            if (len) // There is no "nominal" size of static or
+                     // readonly space, so give them a size
                 spaces[id].desired_size = len;
-            else // Assign some address, so free_pointer does enclose [0 .. addr+0]
+            else { // Assign some address, so free_pointer does enclose [0 .. addr+0]
+                if (id == STATIC_CORE_SPACE_ID)
+                    lose("Static space size is 0?");
                 READ_ONLY_SPACE_START = READ_ONLY_SPACE_END = addr;
+            }
         }
 #endif
         if (len != 0) {
             spaces[id].len = len;
-            // Try to map at address requested by the core file.
             size_t request = spaces[id].desired_size;
             int sub_2gb_flag = (request & 1);
             request &= ~(size_t)1;
+            // Try to map at address requested by the core file unless ASLR.
+#ifdef LISP_FEATURE_ASLR
+            addr = 0;
+#endif
             addr = (uword_t)reserve_space(id, sub_2gb_flag ? MOVABLE_LOW : MOVABLE,
                                           (os_vm_address_t)addr, request);
             switch (id) {
+            case STATIC_CORE_SPACE_ID:
+#ifdef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
+                STATIC_SPACE_START = addr;
+                STATIC_SPACE_END = addr + len;
+#endif
 #ifndef LISP_FEATURE_DARWIN_JIT
             case READ_ONLY_CORE_SPACE_ID:
                 READ_ONLY_SPACE_START = addr;
@@ -832,8 +868,10 @@ process_directory(int count, struct ndir_entry *entry,
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
             case IMMOBILE_FIXEDOBJ_CORE_SPACE_ID:
             case IMMOBILE_TEXT_CORE_SPACE_ID:
+#ifdef LISP_FEATURE_X86_64
                 if (addr + request > 0x80000000)
                     lose("Won't map immobile space above 2GB");
+#endif
                 if (id == IMMOBILE_FIXEDOBJ_CORE_SPACE_ID)
                     FIXEDOBJ_SPACE_START = addr;
                 else
@@ -908,6 +946,11 @@ process_directory(int count, struct ndir_entry *entry,
     set_adjustment(adj, DYNAMIC_SPACE_START, // actual
                    spaces[DYNAMIC_CORE_SPACE_ID].base, // expected
                    spaces[DYNAMIC_CORE_SPACE_ID].len);
+#ifdef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
+    set_adjustment(adj, STATIC_SPACE_START, // actual
+                   spaces[STATIC_CORE_SPACE_ID].base, // expected
+                   spaces[STATIC_CORE_SPACE_ID].len);
+#endif
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     if (lisp_code_in_elf() && TEXT_SPACE_START != spaces[IMMOBILE_TEXT_CORE_SPACE_ID].base) {
         lose("code-in-elf + PIE not supported");
@@ -1046,7 +1089,11 @@ load_core_file(char *file, os_vm_offset_t file_offset, int merge_core_pages)
 char* get_asm_routine_by_name(const char* name, int *index)
 {
     struct code* code = (struct code*)asm_routines_start;
+#ifdef LISP_FEATURE_DARWIN_JIT
     lispobj ht = CONS(code->debug_info)->car;
+#else
+    lispobj ht = code->debug_info;
+#endif
     if (ht) {
         struct vector* table =
             VECTOR(((struct hash_table*)native_pointer(ht))->pairs);
@@ -1078,37 +1125,25 @@ void asm_routine_poke(const char* routine, int offset, char byte)
         address[offset] = byte;
 }
 
-// Caution: use at your own risk
-#if defined DEBUG_CORE_LOADING && DEBUG_CORE_LOADING
-#include "hopscotch.h"
-#include "genesis/cons.h"
-#include "genesis/layout.h"
-#include "genesis/gc-tables.h"
-#include "code.h"
-#include "gc-private.h"
+static void trace_sym(lispobj, struct symbol*, struct grvisit_context*);
 
-struct visitor {
-    // one item per value of widetag>>2
-    // element 0 is for conses, element 64 is for totals.
-    struct {
-        int count;
-        int words;
-    } headers[65], sv_subtypes[3];
-    struct hopscotch_table *reached;
-};
+#define RECURSE(x) if(is_lisp_pointer(x))graph_visit(ptr,x,context)
 
-static void trace_sym(lispobj, struct symbol*, struct hopscotch_table*);
-#define RECURSE(x) if(is_lisp_pointer(x))graph_visit(ptr,x,seen)
-static void graph_visit(lispobj __attribute__((unused)) referer,
-                        lispobj ptr,
-                        struct hopscotch_table* seen)
+/* Despite this being a nice concise expression of a pointer tracing algorithm,
+ * it turns out to be almost unusable in any sufficiently complicated object graph
+ * due to stack overflow. The pristine core alone hits a recursion depth of >8000. */
+static void graph_visit(lispobj referer, lispobj ptr, struct grvisit_context* context)
 {
+#define ILLEGAL_WORD 0xFFFFFFFFDEADBEEF
+    if (ptr == ILLEGAL_WORD) lose("object %"OBJ_FMTX" contains a garbage word", referer);
     if (lowtag_of(ptr) == FUN_POINTER_LOWTAG
-        && widetag_of(FUNCTION(ptr)) == SIMPLE_FUN_WIDETAG)
+        && widetag_of((lispobj*)FUNCTION(ptr)) == SIMPLE_FUN_WIDETAG)
         ptr = fun_code_tagged(FUNCTION(ptr));
-    if (hopscotch_get(seen, ptr, 0))
-        return;
-    hopscotch_insert(seen, ptr, 1);
+    if (hopscotch_get(context->seen, ptr, 0)) return;
+    if (++context->depth > context->maxdepth) context->maxdepth = context->depth;
+    // TODO: add rejection function for off-heap objects as part of supplied context
+    hopscotch_insert(context->seen, ptr, 1);
+    if (context->action) context->action(ptr, context->data);
     lispobj layout, *obj;
     sword_t nwords, i;
     if (lowtag_of(ptr) == LIST_POINTER_LOWTAG) {
@@ -1125,7 +1160,7 @@ static void graph_visit(lispobj __attribute__((unused)) referer,
         case INSTANCE_WIDETAG:
         case FUNCALLABLE_INSTANCE_WIDETAG:
             layout = layout_of(obj);
-            graph_visit(ptr, layout, seen);
+            graph_visit(ptr, layout, context);
             nwords = headerobj_size(obj);
             struct bitmap bitmap = get_layout_bitmap(LAYOUT(layout));
             for (i=0; i<(nwords-1); ++i)
@@ -1141,7 +1176,7 @@ static void graph_visit(lispobj __attribute__((unused)) referer,
         // then we scan word indices 1 and 2 off the object base address.
         case CLOSURE_WIDETAG:
             // We must scan the closure's trampoline word.
-            graph_visit(ptr, fun_taggedptr_from_self(obj[1]), seen);
+            graph_visit(ptr, fun_taggedptr_from_self(obj[1]), context);
             // Closures can utilize one payload word beyond what the header
             // indicates. This is quite sucky and I don't know why I did that.
             // However, it is correctly accounted for by SHORT_BOXED_NWORDS
@@ -1150,11 +1185,7 @@ static void graph_visit(lispobj __attribute__((unused)) referer,
             for(i=2; i<=nwords; ++i) RECURSE(obj[i]);
             break;
         case SYMBOL_WIDETAG:
-            trace_sym(ptr, SYMBOL(ptr), seen);
-            break;
-        case WEAK_POINTER_WIDETAG:
-            nwords = TINY_BOXED_NWORDS(*obj);
-            for(i=1; i<=nwords; ++i) RECURSE(obj[i]);
+            trace_sym(ptr, SYMBOL(ptr), context);
             break;
         case FDEFN_WIDETAG:
             RECURSE(obj[1]);
@@ -1162,13 +1193,17 @@ static void graph_visit(lispobj __attribute__((unused)) referer,
             RECURSE(decode_fdefn_rawfun((struct fdefn*)obj));
             break;
         default:
+            // weak-pointer can be considered an ordinary boxed object.
+            // the 'next' link looks like a fixnum.
             if (!leaf_obj_widetag_p(widetag_of(obj))) {
                 sword_t size = headerobj_size(obj);
                 for(i=1; i<size; ++i) RECURSE(obj[i]);
             }
       }
+      --context->depth;
 }
-static void trace_sym(lispobj ptr, struct symbol* sym, struct hopscotch_table* seen)
+
+static void trace_sym(lispobj ptr, struct symbol* sym, struct grvisit_context* context)
 {
     RECURSE(decode_symbol_name(sym->name));
     RECURSE(sym->value);
@@ -1176,14 +1211,57 @@ static void trace_sym(lispobj ptr, struct symbol* sym, struct hopscotch_table* s
     RECURSE(sym->fdefn);
 }
 
+/* Caller must provide an uninitialized hopscotch table.
+ * This function will initialize it and perform a graph visit.
+ * Caller may subsequently inspect the table and/or visit other objects as
+ * dictated by thread stacks, etc. Caller may - but need not - provide
+ * an 'action' to invoke on each object */
+struct grvisit_context*
+visit_heap_from_static_roots(struct hopscotch_table* reached,
+                             void (*action)(lispobj, void*),
+                             void* data)
+{
+    hopscotch_create(reached, HOPSCOTCH_HASH_FUN_DEFAULT,
+                     0, // no values
+                     1<<18, /* initial size */
+                     0);
+
+    struct grvisit_context* context = malloc(sizeof (struct grvisit_context));
+    context->seen = reached;
+    context->action = action;
+    context->data = data;
+    context->depth = context->maxdepth = 0;
+    trace_sym(NIL, SYMBOL(NIL), context);
+    lispobj* where = (lispobj*)STATIC_SPACE_OBJECTS_START;
+    lispobj* end = static_space_free_pointer;
+    while (where<end) {
+        graph_visit(0, compute_lispobj(where), context);
+        where += object_size(where);
+    }
+    return context;
+}
+
+// Caution: use at your own risk
+#if defined DEBUG_CORE_LOADING && DEBUG_CORE_LOADING
+struct visitor {
+    // one item per value of widetag>>2
+    // element 0 is for conses, element 64 is for totals.
+    struct {
+        int count;
+        int words;
+    } headers[65], sv_subtypes[3];
+    struct hopscotch_table *reached;
+};
+
 static void tally(lispobj ptr, struct visitor* v)
 {
+    sword_t words;
     if (lowtag_of(ptr) == LIST_POINTER_LOWTAG)
-        ++v->headers[0].count;
+        ++v->headers[0].count, words = 2;
     else {
         lispobj* obj = native_pointer(ptr);
         lispobj header = *obj;
-        sword_t words = object_size2(obj, header);
+        words = object_size2(obj, header);
         int widetag = header_widetag(header);
         int header_index = widetag>>2;
         ++v->headers[header_index].count;
@@ -1236,25 +1314,16 @@ static void sanity_check_loaded_core(lispobj initial_function)
     struct hopscotch_table reached;
     memset(v, 0, sizeof v);
     // Pass 1: Count objects reachable from known roots.
-    hopscotch_create(&reached, HOPSCOTCH_HASH_FUN_DEFAULT,
-                     0, // no values
-                     1<<18, /* initial size */
-                     0);
-    {
-      trace_sym(NIL, SYMBOL(NIL), &reached);
-      lispobj* where = (lispobj*)STATIC_SPACE_OBJECTS_START;
-      lispobj* end = static_space_free_pointer;
-      while (where<end) {
-        graph_visit(0, compute_lispobj(where), &reached);
-        where += object_size(where);
-      }
-    }
-    graph_visit(0, initial_function, &reached); // not otherwise reachable
+    struct grvisit_context* c
+        = visit_heap_from_static_roots(&reached, 0, 0);
+    graph_visit(0, initial_function, c); // initfun is not otherwise reachable
     // having computed the reaching graph, tally up the dynamic space objects
     int key_index;
     lispobj ptr;
     for_each_hopscotch_key(key_index, ptr, reached)
         if (count_this_pointer_p(ptr)) tally(ptr, &v[0]);
+    printf("graphvisit.maxdepth=%d\n", c->maxdepth);
+    free(c);
     // Pass 2: Count all heap objects
     v[1].reached = &reached;
     walk_generation(visit, -1, (uword_t)&v[1]);
