@@ -38,7 +38,7 @@
 #include "genesis/primitive-objects.h"
 #include "genesis/gc-tables.h"
 #include "gc-internal.h"
-#include "tlsf-bsd/tlsf/tlsf.h"
+#include "../../tlsf-bsd/tlsf/tlsf.h"
 extern void* tlsf_control;
 
 /* When we need to do command input, we use this stream, which is not
@@ -130,12 +130,12 @@ void save_gc_crashdump(char *pathname,
     preamble.pin_dynspace_code = pin_all_dynamic_space_code;
     preamble.sizeof_context = sizeof (os_context_t);
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
-    char *tlsf_memory_end = (char*)TEXT_SPACE_START + TEXT_SPACE_SIZE;
+    char *tlsf_memory_end = (char*)TEXT_SPACE_START + text_space_size;
     preamble.fixedobj_start = FIXEDOBJ_SPACE_START;
     preamble.fixedobj_size = FIXEDOBJ_SPACE_SIZE;
     preamble.fixedobj_free_pointer = (uword_t)fixedobj_free_pointer;
     preamble.text_start = TEXT_SPACE_START;
-    preamble.text_size = TEXT_SPACE_SIZE;
+    preamble.text_size = text_space_size;
     preamble.text_space_highwatermark = text_space_highwatermark;
     preamble.tlsf_mem_start = tlsf_mem_start;
     preamble.tlsf_control_address = tlsf_control;
@@ -159,7 +159,7 @@ void save_gc_crashdump(char *pathname,
     usage = (uword_t)text_space_highwatermark - TEXT_SPACE_START;
     // write the block_header_t that is just beyond the high water mark
     checked_write("text", fd, (char*)TEXT_SPACE_START, usage+3*N_WORD_BYTES);
-    total_npages = TEXT_SPACE_SIZE / IMMOBILE_CARD_BYTES;
+    total_npages = text_space_size / IMMOBILE_CARD_BYTES;
     int n_bitmap_elts = ALIGN_UP(total_npages, 32) / 32;
     checked_write("text_gen", fd, text_page_genmask, total_npages); // 1 byte per page
     checked_write("text_WP", fd, text_page_touched_bits, n_bitmap_elts * sizeof (int));
@@ -233,7 +233,7 @@ static cmd call_cmd, dump_cmd, print_cmd, quit_cmd, help_cmd;
 static cmd flush_cmd, regs_cmd, exit_cmd;
 static cmd print_context_cmd, pte_cmd, search_cmd;
 static cmd backtrace_cmd, catchers_cmd;
-static cmd threads_cmd;
+static cmd threads_cmd, findpath_cmd;
 
 extern void gc_stop_the_world(), gc_start_the_world();
 static void suspend_other_threads() {
@@ -270,11 +270,11 @@ void list_lisp_threads(int regions) {
     for_each_thread(th) {
         memcpy(&pthread, &th->os_thread, N_WORD_BYTES);
         struct thread_instance* i = (void*)(th->lisp_thread - INSTANCE_POINTER_LOWTAG);
-        char* name = NULL;
-        if (i->name != NIL &&
-            widetag_of(native_pointer(i->name)) == SIMPLE_BASE_STRING_WIDETAG)
-            name = (char*)VECTOR(i->name)->data;
-        fprintf(stderr, "%p %p %p \"%s\"\n", th, pthread, (void*)i, name);
+        lispobj name = instancep(th->lisp_thread) ? i->_name : 0;
+        char* cname = NULL;
+        // it's OK to call widetag_of on NIL but not NULL
+        if (name && simple_base_string_p(name)) cname = vector_sap(name);
+        fprintf(stderr, "%p %p %p \"%s\"\n", th, pthread, (void*)i, cname);
         if (regions) {
 #define show_tlab(label, r) fprintf(stderr, "  %s @ %p: %p %p %p\n", label, \
          &th->r, th->r.start_addr, th->r.end_addr, th->r.free_pointer)
@@ -285,11 +285,74 @@ void list_lisp_threads(int regions) {
 #undef show_tlab
         }
     }
+    if (regions) {
+#define show_tlab(label, r) fprintf(stderr, "  %s @ %p: %p %p %p\n", label, \
+         r, r->start_addr, r->end_addr, r->free_pointer)
+        fprintf(stderr, "global regions:\n");
+        show_tlab("mixed    ", mixed_region);
+        show_tlab("smallmix ", small_mixed_region);
+        show_tlab("unboxed  ", unboxed_region);
+        show_tlab("code     ", code_region);
+        show_tlab("boxed    ", boxed_region);
+        show_tlab("cons     ", cons_region);
+#undef show_tlab
+    }
 }
 static int threads_cmd(char **ptr) {
     list_lisp_threads(more_p(ptr) && !strncmp(*ptr, "-r", 2));
     return 0;
 }
+extern int heap_trace_verbose;
+extern int gc_pathfind_aux(lispobj*, lispobj, lispobj, lispobj, int);
+
+static int findpath_cmd(char **ptr) {
+    // prevent the path finder from seeing the object that results from parsing the command
+    lispobj* stackptr = (lispobj*)&ptr;
+    // overaligned for 32-bit but doesn't matter
+    struct vector __attribute__ ((aligned (16))) result;
+    struct cons __attribute__ ((aligned (16))) list;
+    /* Despite capturing the stack pointer coming in to this function,
+     * it's very difficult to ensure that a locally allocated weak pointer
+     * is not above that (and hence its value slot seen).
+     * So it has to be malloc()ed. */
+    char* wp_mem = malloc(sizeof (struct weak_pointer) + N_WORD_BYTES);
+    struct weak_pointer* wp =
+      (void*)(wp_mem + (((uword_t)wp_mem & LOWTAG_MASK) ? N_WORD_BYTES : 0));
+    wp->header = ((WEAK_POINTER_SIZE-1)<<N_WIDETAG_BITS)|WEAK_POINTER_WIDETAG;
+    if (parse_lispobj(ptr, &wp->value)) {
+        list.car = make_lispobj(wp, OTHER_POINTER_LOWTAG);
+        list.cdr = NIL;
+        result.header = SIMPLE_VECTOR_WIDETAG;
+        result.length_ = make_fixnum(1);
+        result.data[0] = 0;
+        suspend_other_threads();
+        int save = heap_trace_verbose;
+        heap_trace_verbose = 4;
+        gc_pathfind_aux(stackptr,
+                        make_lispobj(&list, LIST_POINTER_LOWTAG),
+                        make_lispobj(&result, OTHER_POINTER_LOWTAG),
+                        0, 2);
+        heap_trace_verbose = save;
+        unsuspend_other_threads();
+        lispobj path = result.data[0];
+        if (listp(path)) {
+            fprintf(stderr, "Answer:\n");
+            while (path != NIL) {
+                struct cons* pair = CONS(CONS(path)->car);
+                if (listp(pair->cdr)) {
+                    // thread root - complicated to print
+                } else {
+                    // otherwise, object and word index
+                    fprintf(stderr, " %"OBJ_FMTX" word %d\n",
+                            pair->car, (int)pair->cdr);
+                }
+                path = CONS(path)->cdr;
+            }
+        }
+    }
+    free(wp_mem);
+    return 0;
+}
 static int verify_cmd(char __attribute__((unused)) **ptr) {
     gencgc_verbose = 1;
     suspend_other_threads();
@@ -300,7 +363,6 @@ static int verify_cmd(char __attribute__((unused)) **ptr) {
 static int gc_cmd(char **ptr) {
     int last_gen = 0;
     extern generation_index_t verify_gens;
-    extern boolean pre_verify_gen_0;
     if (more_p(ptr)) parse_number(ptr, &last_gen);
     gencgc_verbose = 2;
     pre_verify_gen_0 = 1;
@@ -314,6 +376,10 @@ static int gc_cmd(char **ptr) {
 static int tlsf_cmd(__attribute__((unused)) char **ptr) {
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     tlsf_dump_pool(tlsf_control, tlsf_mem_start, "/dev/tty");
+#ifdef TLSF_CONFIG_DEBUG
+    tlsf_check(tlsf_control);
+    tlsf_check_pool(tlsf_mem_start);
+#endif
 #endif
     return 0;
 }
@@ -331,6 +397,7 @@ static struct cmd {
     {"dump", "Dump memory starting at ADDRESS for COUNT words.", dump_cmd},
     {"d", "(an alias for dump)", dump_cmd},
     {"exit", "Exit this instance of the monitor.", exit_cmd},
+    {"findpath", "Find path to an object.", findpath_cmd},
     {"flush", "Flush all temp variables.", flush_cmd},
     {"print", "Print object at ADDRESS.", print_cmd},
     {"p", "(an alias for print)", print_cmd},
@@ -355,7 +422,7 @@ visible(unsigned char c)
         return c;
 }
 
-static boolean valid_widetag_p(unsigned char widetag) {
+static bool valid_widetag_p(unsigned char widetag) {
     // TODO: ensure that widetag is defined (not "unused") and is for a headered object
     // (i.e. is not CHARACTER_WIDETAG and not some other things)
     return other_immediate_lowtag_p(widetag);
@@ -399,7 +466,7 @@ dump_cmd(char **ptr)
         count = -count;
     }
 
-    boolean aligned = ((uword_t)addr & LOWTAG_MASK) == 0;
+    bool aligned = ((uword_t)addr & LOWTAG_MASK) == 0;
     if (decode && (!aligned || displacement < 0)) {
         printf("Sorry, can only decode if aligned and stepping forward\n");
         decode = 0;
@@ -485,7 +552,7 @@ pte_cmd(char **ptr)
 static int
 regs_cmd(char __attribute__((unused)) **ptr)
 {
-    struct thread __attribute__((unused)) *thread=get_sb_vm_thread();
+    struct thread __attribute__((unused)) *thread = get_sb_vm_thread();
 
     printf("CSP\t=\t%p   ", access_control_stack_pointer(thread));
 #if !defined(LISP_FEATURE_X86) && !defined(LISP_FEATURE_X86_64)
@@ -645,7 +712,7 @@ static int
 print_context_cmd(char **ptr)
 {
     int free_ici;
-    struct thread *thread=get_sb_vm_thread();
+    struct thread *thread = get_sb_vm_thread();
 
     free_ici = fixnum_value(read_TLS(FREE_INTERRUPT_CONTEXT_INDEX,thread));
 
@@ -729,7 +796,6 @@ catchers_cmd(char __attribute__((unused)) **ptr)
     return 0;
 }
 
-extern boolean gc_active_p;
 extern FILE *gc_activitylog_file;
 void
 ldb_monitor(void)
@@ -793,13 +859,6 @@ ldb_monitor(void)
     }
 }
 
-/* what we do when things go badly wrong at a low level */
-void
-monitor_or_something()
-{
-    ldb_monitor();
-}
-
 #ifdef STANDALONE_LDB
 void gc_stop_the_world() { } // do nothing
 void gc_start_the_world() { } // do nothing
@@ -810,23 +869,23 @@ struct lisp_startup_options lisp_startup_options;
 
 void unwind_binding_stack() { lose("Can't unwind binding stack"); }
 FILE *prepare_to_save(__attribute__((unused)) char *filename,
-                      __attribute__((unused)) boolean prepend_runtime,
+                      __attribute__((unused)) bool prepend_runtime,
                       __attribute__((unused)) void **runtime_bytes,
                       __attribute__((unused)) size_t *runtime_size) {
     lose("Can't prepare_to_save");
 }
-boolean save_runtime_to_filehandle(__attribute__((unused)) FILE *output,
-                                   __attribute__((unused)) void *runtime,
-                                   __attribute__((unused)) size_t runtime_size,
-                                   __attribute__((unused)) int application_type) {
+bool save_runtime_to_filehandle(__attribute__((unused)) FILE *output,
+                                __attribute__((unused)) void *runtime,
+                                __attribute__((unused)) size_t runtime_size,
+                                __attribute__((unused)) int application_type) {
     lose("Can't save_runtime_to_filehandle");
 }
-boolean save_to_filehandle(__attribute__((unused)) FILE *file,
-                           __attribute__((unused)) char *filename,
-                           __attribute__((unused)) lispobj init_function,
-                           __attribute__((unused)) boolean make_executable,
-                           __attribute__((unused)) boolean save_runtime_options,
-                           __attribute__((unused)) int core_compression_level) {
+bool save_to_filehandle(__attribute__((unused)) FILE *file,
+                        __attribute__((unused)) char *filename,
+                        __attribute__((unused)) lispobj init_function,
+                        __attribute__((unused)) bool make_executable,
+                        __attribute__((unused)) bool save_runtime_options,
+                        __attribute__((unused)) int core_compression_level) {
     lose("Can't save_to_filehandle");
 }
 
@@ -916,7 +975,7 @@ int load_gc_crashdump(char* pathname)
                 calc_immobile_space_bounds(),
                 write_protect_immobile_space();
     gc_assert(preamble.fixedobj_size == FIXEDOBJ_SPACE_SIZE);
-    gc_assert(preamble.text_size == TEXT_SPACE_SIZE);
+    gc_assert(preamble.text_size == text_space_size);
     FIXEDOBJ_SPACE_START = preamble.fixedobj_start;
     TEXT_SPACE_START = preamble.text_start;
     fixedobj_free_pointer = (lispobj*)preamble.fixedobj_free_pointer;
@@ -924,7 +983,7 @@ int load_gc_crashdump(char* pathname)
     os_alloc_gc_space(IMMOBILE_FIXEDOBJ_CORE_SPACE_ID, 0, (char*)FIXEDOBJ_SPACE_START,
                       FIXEDOBJ_SPACE_SIZE);
     os_alloc_gc_space(IMMOBILE_TEXT_CORE_SPACE_ID, 0, (char*)TEXT_SPACE_START,
-                      TEXT_SPACE_SIZE);
+                      text_space_size);
     gc_init_immobile(); // allocate the page tables
     calc_immobile_space_bounds();
     // Read fixedobj space
@@ -935,13 +994,13 @@ int load_gc_crashdump(char* pathname)
     checked_read("fixedobj_PTE", fd, fixedobj_pages, total_npages * sizeof sizeof(struct fixedobj_page));
     // Read text space
     usage = (uword_t)text_space_highwatermark - TEXT_SPACE_START;
-    char *tlsf_memory_end = (char*)TEXT_SPACE_START + TEXT_SPACE_SIZE;
+    char *tlsf_memory_end = (char*)TEXT_SPACE_START + text_space_size;
     tlsf_mem_start = preamble.tlsf_mem_start;
     fprintf(stderr, "tlsf_mem_start=%p\n", tlsf_mem_start);
     int tlsf_memory_size = tlsf_memory_end - (char*)tlsf_mem_start;
     checked_read("text", fd, (char*)TEXT_SPACE_START, usage+3*N_WORD_BYTES);
     memcpy(tlsf_memory_end-3*N_WORD_BYTES, preamble.sentinel_block, 3*N_WORD_BYTES);
-    total_npages = TEXT_SPACE_SIZE / IMMOBILE_CARD_BYTES;
+    total_npages = text_space_size / IMMOBILE_CARD_BYTES;
     int n_bitmap_elts = ALIGN_UP(total_npages, 32) / 32;
     checked_read("text_gen", fd, text_page_genmask, total_npages); // 1 byte per page
     checked_read("text_WP", fd, text_page_touched_bits, n_bitmap_elts * sizeof (int));
@@ -989,10 +1048,9 @@ int load_gc_crashdump(char* pathname)
         checked_read(" TLS", fd, &th->lisp_thread, preamble.tls_size-skip);
         write_TLS(FREE_INTERRUPT_CONTEXT_INDEX, make_fixnum(1), th);
         struct thread_instance* instance = (void*)(th->lisp_thread - INSTANCE_POINTER_LOWTAG);
-        lispobj name = instance->name;
-        char* cname = (gc_managed_addr_p(name) &&
-                       widetag_of(native_pointer(name)) == SIMPLE_BASE_STRING_WIDETAG)
-                      ? (char*)name+1 : 0;
+        lispobj name = instance->_name;
+        char* cname = gc_managed_addr_p(name) && simple_base_string_p(name)
+                      ? vector_sap(name) : 0;
         fprintf(stderr, "thread @ %p originally %p, %d bind_stk words, %d val_stk words '%s'\n",
                 th, (void*)thread_preamble.address,
                 (int)(thread_preamble.binding_stack_nbytes>>WORD_SHIFT),
@@ -1045,7 +1103,7 @@ int main(int argc, char *argv[], char **envp)
         fprintf(stderr, "Usage: ldb crashdump\n");
         return 1;
     }
-    boolean have_hardwired_spaces = os_preinit(argv, envp);
+    bool have_hardwired_spaces = os_preinit(argv, envp);
     allocate_lisp_dynamic_space(have_hardwired_spaces);
     gc_init();
     load_gc_crashdump(argv[1]);
