@@ -20,8 +20,7 @@
 #include "os.h"
 #include "arch.h"
 #include "lispregs.h"
-#include "signal.h"
-#include "alloc.h"
+#include <signal.h>
 #include "interrupt.h"
 #include "interr.h"
 #include "breakpoint.h"
@@ -73,21 +72,6 @@ static void xgetbv(unsigned *eax, unsigned *edx)
 
 #define VECTOR_FILL_T "VECTOR-FILL/T"
 
-void set_alloc_tramp_vectors(int ymm_enable)
-{
-    struct code* code = (struct code*)asm_routines_start;
-    lispobj* instructions = (lispobj*)code + code_header_words(code);
-    int index;
-    get_asm_routine_by_name("FPR-SAVE", &index);
-    if (index)
-        instructions[index] =
-            (lispobj)get_asm_routine_by_name(ymm_enable ? "SAVE-YMM" : "SAVE-XMM", 0);
-    get_asm_routine_by_name("FPR-RESTORE", &index);
-    if (index)
-        instructions[index] =
-            (lispobj)get_asm_routine_by_name(ymm_enable ? "RESTORE-YMM" : "RESTORE-XMM", 0);
-}
-
 // Poke in a byte that changes an opcode to enable faster vector fill.
 // Using fixed offsets and bytes is no worse than what we do elsewhere.
 void tune_asm_routines_for_microarch(void)
@@ -111,7 +95,6 @@ void tune_asm_routines_for_microarch(void)
             }
         }
     }
-    set_alloc_tramp_vectors(avx_supported);
     int our_cpu_feature_bits = 0;
     // avx2_supported gets copied into bit 1 of *CPU-FEATURE-BITS*
     if (avx2_supported) our_cpu_feature_bits |= 1;
@@ -138,7 +121,6 @@ void untune_asm_routines_for_microarch(void)
 {
     asm_routine_poke(VECTOR_FILL_T, 0x12, 0xEB); // Change JL to JMP
     SetSymbolValue(CPU_FEATURE_BITS, 0, 0);
-    set_alloc_tramp_vectors(0);
 }
 
 #ifndef _WIN64
@@ -159,6 +141,30 @@ arch_get_bad_addr(int __attribute__((unused)) sig,
  * want to get to, and on OS, which determines how we get to it.)
  */
 
+// I don't have an easy way to test changes for these OSes, so just use this
+// context visitor based on the old slightly-less-efficient way of doing it.
+#if defined LISP_FEATURE_SUNOS || defined LISP_FEATURE_HAIKU
+void visit_context_registers(void (*proc)(os_context_register_t, void*),
+                             os_context_t *context, void* arg)
+{
+    proc(os_context_pc(context), arg);
+    proc(*os_context_register_addr(context, reg_RAX), arg);
+    proc(*os_context_register_addr(context, reg_RCX), arg);
+    proc(*os_context_register_addr(context, reg_RDX), arg);
+    proc(*os_context_register_addr(context, reg_RBX), arg);
+    proc(*os_context_register_addr(context, reg_RSI), arg);
+    proc(*os_context_register_addr(context, reg_RDI), arg);
+    proc(*os_context_register_addr(context, reg_R8 ), arg);
+    proc(*os_context_register_addr(context, reg_R9 ), arg);
+    proc(*os_context_register_addr(context, reg_R10), arg);
+    proc(*os_context_register_addr(context, reg_R11), arg);
+    proc(*os_context_register_addr(context, reg_R12), arg);
+    proc(*os_context_register_addr(context, reg_R13), arg);
+    proc(*os_context_register_addr(context, reg_R14), arg);
+    proc(*os_context_register_addr(context, reg_R15), arg);
+}
+#endif
+
 os_context_register_t *
 os_context_flags_addr(os_context_t *context)
 {
@@ -231,8 +237,6 @@ void arch_skip_instruction(os_context_t *context)
             break;
         }
 
-    FSHOW((stderr,
-           "/[arch_skip_inst resuming at %x]\n", OS_CONTEXT_PC(context)));
 }
 
 unsigned char *
@@ -241,23 +245,15 @@ arch_internal_error_arguments(os_context_t *context)
     return 1 + (unsigned char *)OS_CONTEXT_PC(context);
 }
 
-boolean
-arch_pseudo_atomic_atomic(os_context_t __attribute__((unused)) *context)
-{
-    return get_pseudo_atomic_atomic(get_sb_vm_thread());
+bool arch_pseudo_atomic_atomic(struct thread *thread) {
+    return get_pseudo_atomic_atomic(thread);
 }
 
-void
-arch_set_pseudo_atomic_interrupted(os_context_t __attribute__((unused)) *context)
-{
-    struct thread *thread = get_sb_vm_thread();
+void arch_set_pseudo_atomic_interrupted(struct thread *thread) {
     set_pseudo_atomic_interrupted(thread);
 }
 
-void
-arch_clear_pseudo_atomic_interrupted(os_context_t __attribute__((unused)) *context)
-{
-    struct thread *thread = get_sb_vm_thread();
+void arch_clear_pseudo_atomic_interrupted(struct thread *thread) {
     clear_pseudo_atomic_interrupted(thread);
 }
 
@@ -584,7 +580,7 @@ arch_set_fp_modes(unsigned int mxcsr)
     asm ("ldmxcsr %0" : : "m" (temp));
 }
 
-static __attribute__((unused)) boolean codeblob_p(lispobj ptr) {
+static __attribute__((unused)) bool codeblob_p(lispobj ptr) {
     return lowtag_of(ptr) == OTHER_POINTER_LOWTAG &&
            widetag_of((lispobj*)(ptr-OTHER_POINTER_LOWTAG)) == CODE_HEADER_WIDETAG;
 }
@@ -649,26 +645,26 @@ lispobj decode_fdefn_rawfun(struct fdefn* fdefn) {
     return entrypoint_taggedptr((uword_t)fdefn->raw_addr);
 }
 
+#ifdef LISP_FEATURE_SB_THREAD
 #include "genesis/vector.h"
 #define LOCK_PREFIX 0xF0
 #undef SHOW_PC_RECORDING
 
-extern unsigned int alloc_profile_n_counters;
 extern unsigned int max_alloc_point_counters;
-#ifdef LISP_FEATURE_SB_THREAD
+
 #ifdef LISP_FEATURE_WIN32
 extern CRITICAL_SECTION alloc_profiler_lock;
 #else
 extern pthread_mutex_t alloc_profiler_lock;
 #endif
-#endif
 
-static unsigned int claim_index(int qty)
+
+static unsigned int claim_index(int qty) // qty is 1 or 2
 {
-    static boolean warning_issued;
-    unsigned int index = alloc_profile_n_counters;
-    alloc_profile_n_counters += qty;
-    if (alloc_profile_n_counters <= max_alloc_point_counters)
+    static bool warning_issued;
+    unsigned int index = fixnum_value(SYMBOL(N_PROFILE_SITES)->value);
+    SYMBOL(N_PROFILE_SITES)->value += make_fixnum(qty);
+    if (fixnum_value(SYMBOL(N_PROFILE_SITES)->value) <= max_alloc_point_counters)
        return index;
     if (!warning_issued) {
        fprintf(stderr, "allocation profile buffer overflowed\n");
@@ -677,7 +673,7 @@ static unsigned int claim_index(int qty)
     return 0; // use the overflow bin(s)
 }
 
-static boolean NO_SANITIZE_MEMORY
+static bool NO_SANITIZE_MEMORY
 instrumentp(uword_t* sp, uword_t** pc, uword_t* old_word)
 {
     int __attribute__((unused)) ret = mutex_acquire(&alloc_profiler_lock);
@@ -705,7 +701,7 @@ instrumentp(uword_t* sp, uword_t** pc, uword_t* old_word)
 
 // logical index 'index' in the metadata array stores the code component
 // and pc-offset relative to the component base address
-static void record_pc(char* pc, unsigned int index, boolean sizedp)
+static void record_pc(char* pc, unsigned int index, bool sizedp)
 {
     lispobj *code = component_ptr_from_pc(pc);
     if (!code) {
@@ -747,8 +743,7 @@ static void record_pc(char* pc, unsigned int index, boolean sizedp)
     }
 }
 
-void AMD64_SYSV_ABI
-allocation_tracker_counted(uword_t* sp)
+void allocation_tracker_counted(uword_t* sp)
 {
     uword_t *pc, word_at_pc;
     if (instrumentp(sp, &pc, &word_at_pc)) {
@@ -779,8 +774,7 @@ allocation_tracker_counted(uword_t* sp)
     gc_assert(ret);
 }
 
-void AMD64_SYSV_ABI
-allocation_tracker_sized(uword_t* sp)
+void allocation_tracker_sized(uword_t* sp)
 {
     uword_t *pc, word_at_pc;
     if (instrumentp(sp, &pc, &word_at_pc)) {
@@ -811,6 +805,7 @@ allocation_tracker_sized(uword_t* sp)
     int __attribute__((unused)) ret = mutex_release(&alloc_profiler_lock);
     gc_assert(ret);
 }
+#endif
 
 __attribute__((sysv_abi)) lispobj call_into_lisp(lispobj fun, lispobj *args, int nargs) {
     extern lispobj call_into_lisp_(lispobj, lispobj *, int, struct thread *)
